"""
Created on 07 Apr 2021

@author: Jade Page (jade.page@southcoastscience.com)

https://stackoverflow.com/questions/36932/how-can-i-represent-an-enum-in-python
"""

from collections import OrderedDict
from enum import Enum
from http import HTTPStatus

from scs_core.aws.data.http_response import HTTPResponse
from scs_core.data.str import Str
from scs_core.sample.configuration_sample import ConfigurationSample
from scs_core.sys.http_exception import HTTPException


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationFinder(object):
    """
    classdocs
    """

    __URL = "https://p18hyi3w56.execute-api.us-west-2.amazonaws.com/default/ConfigurationFinder"

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, http_client, auth):
        self.__http_client = http_client
        self.__auth = auth


    # ----------------------------------------------------------------------------------------------------------------

    def find(self, tag_filter, exact_match, response_mode):
        request = ConfigurationRequest(tag_filter, exact_match, response_mode)
        headers = {'Authorization': self.__auth.email_address}

        response = self.__http_client.get(self.__URL, headers=headers, params=request.params())

        return ConfigurationResponse.construct_from_jdict(response.json())


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationFinder:{auth:%s}" % self.__auth


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationRequest(object):
    """
    classdocs
    """

    MODE = Enum('Mode', 'LATEST HISTORY DIFF TAGS_ONLY')

    TAG_FILTER = 'tagFilter'
    EXACT_MATCH = 'exactMatch'
    RESPONSE_MODE = 'responseMode'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_qsp(cls, qsp):
        if not qsp:
            return None

        tag_filter = qsp.get(cls.TAG_FILTER)
        exact_match = qsp.get(cls.EXACT_MATCH, 'false').lower() == 'true'

        try:
            response_mode = cls.MODE[qsp.get(cls.RESPONSE_MODE)]
        except KeyError:
            response_mode = None

        return cls(tag_filter, exact_match, response_mode)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag_filter, exact_match, response_mode):
        """
        Constructor
        """
        self.__tag_filter = tag_filter                          # string
        self.__exact_match = bool(exact_match)                  # bool
        self.__response_mode = response_mode                    # MODE enum


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.response_mode is None:
            return False

        return True


    def latest(self):
        return self.__response_mode == self.MODE.LATEST


    def history(self):
        return self.__response_mode == self.MODE.HISTORY


    def diff(self):
        return self.__response_mode == self.MODE.DIFF


    def tags_only(self):
        return self.__response_mode == self.MODE.TAGS_ONLY


    # ----------------------------------------------------------------------------------------------------------------

    def params(self):
        params = {
            self.TAG_FILTER: self.tag_filter,
            self.EXACT_MATCH: self.exact_match,
            self.RESPONSE_MODE: self.response_mode.name
        }

        return params

    # ----------------------------------------------------------------------------------------------------------------

    @property
    def tag_filter(self):
        return self.__tag_filter


    @property
    def exact_match(self):
        return self.__exact_match


    @property
    def response_mode(self):
        return self.__response_mode


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationRequest:{tag_filter:%s, exact_match:%s, response_mode:%s}" % \
               (self.tag_filter, self.exact_match, self.response_mode)


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationResponse(HTTPResponse):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        # print("jdict: %s" % jdict)

        status = HTTPStatus(jdict.get('statusCode'))

        if status != HTTPStatus.OK:
            raise HTTPException(status.value, status.phrase, status.description)

        mode = ConfigurationRequest.MODE[jdict.get('mode')]

        items = []
        if jdict.get('Items'):
            for item_jdict in jdict.get('Items'):
                item = item_jdict.get('tag') if mode == ConfigurationRequest.MODE.TAGS_ONLY else \
                    ConfigurationSample.construct_from_jdict(item_jdict)
                items.append(item)

        next_url = jdict.get('next')

        return cls(status, mode, items, next_url=next_url)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, status, mode, items, next_url=None):
        """
        Constructor
        """
        super().__init__(status)

        self.__mode = mode                                  # ConfigurationRequest.Mode member
        self.__items = items                                # list of ConfigurationSample or string
        self.__next_url = next_url                          # URL string


    def __len__(self):
        return len(self.items)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['statusCode'] = self.status.value

        if self.mode is not None:
            jdict['mode'] = self.mode.name

        if self.items is not None:
            jdict['Items'] = self.items
            jdict['itemCount'] = len(self.items)

        if self.next_url is not None:
            jdict['next'] = self.next_url

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def mode(self):
        return self.__mode


    @property
    def items(self):
        return self.__items


    @property
    def next_url(self):
        return self.__next_url


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationResponse:{status:%s, mode:%s, items:%s, next_url:%s}" % \
               (self.status, self.mode, Str.collection(self.items), self.next_url)



"""
Created on 15 Oct 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
{"CO": {"calibrated-on": "2021-01-19T10:07:27Z", "offset": 2, "env": {"hmd": 41.5, "tmp": 22.1, "pA": null}},
"NO2": {"calibrated-on": "2021-01-19T11:07:27Z", "offset": 1, "env": {"hmd": 41.5, "tmp": 22.1, "pA": null}}}
"""

from abc import ABC
from collections import OrderedDict

from scs_core.data.json import PersistentJSONable

from scs_core.gas.sensor_baseline import SensorBaseline


# --------------------------------------------------------------------------------------------------------------------

class Baseline(PersistentJSONable, ABC):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls({}) if skeleton else None

        sensor_baselines = {}

        for gas, baseline_jdict in jdict.items():
            sensor_baselines[gas] = SensorBaseline.construct_from_jdict(baseline_jdict)

        return cls(sensor_baselines)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, sensor_baselines):
        """
        Constructor
        """
        super().__init__()

        self.__sensor_baselines = sensor_baselines          # dict of gas: SensorBaseline


    def __eq__(self, other):
        try:
            if list(self.gases()) != list(other.gases()):
                return False

            for gas in self.gases():
                if self.sensor_baseline(gas) != other.sensor_baseline(gas):
                    return False

            return True

        except (TypeError, AttributeError):
            return False


    def __len__(self):
        return len(self.__sensor_baselines)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        for gas in sorted(self.__sensor_baselines.keys()):
            jdict[gas] = self.__sensor_baselines[gas]

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def gases(self):
        return self.__sensor_baselines.keys()


    def offsets(self, gases=None):
        return {gas: self.sensor_offset(gas) for gas in (self.gases() if gases is None else gases)}


    def sensor_offset(self, gas):
        baseline = self.sensor_baseline(gas)
        return 0 if baseline is None else baseline.offset


    def sensor_baseline(self, gas):
        try:
            return self.__sensor_baselines[gas]
        except KeyError:
            return None


    def set_sensor_baseline(self, gas, sensor_baseline):
        self.__sensor_baselines[gas] = sensor_baseline


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def sensor_baselines(self):
        return self.__sensor_baselines



"""
Created on 14 Oct 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

A low-performance linear regression utility that does not require scipy

WARNING: if used with time_relative=False and the data time intervals are small, a division by zero error may result
because of the precision limitations of Python float.
"""

from statistics import mean

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.regression import Regression


# --------------------------------------------------------------------------------------------------------------------

class LinearRegression(Regression):
    """
    classdocs
    """

    MIN_DATA_POINTS =   2

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tally=None, time_relative=True):
        """
        Constructor
        """
        self.__tally = tally                            # number of rolling samples (None for all samples)
        self.__time_relative = time_relative            # set first timestamp to time zero

        self.__start_timestamp = None
        self.__tzinfo = None
        self.__data = []


    def __len__(self):
        return len(self.__data)


    # ----------------------------------------------------------------------------------------------------------------

    def has_midpoint(self):
        return len(self) > 0


    def has_regression(self):
        return len(self) >= self.MIN_DATA_POINTS


    def append(self, rec: LocalizedDatetime, value):
        count = len(self)

        if count == 0:
            self.__start_timestamp = rec.timestamp()

        timestamp = rec.timestamp() - self.__start_timestamp if self.__time_relative else rec.timestamp()

        # remove oldest?
        if self.__tally is not None and count == self.__tally:
            del self.__data[0]

        # append...
        self.__data.append((timestamp, value))

        self.__tzinfo = rec.tzinfo


    def reset(self):
        self.__start_timestamp = None
        self.__tzinfo = None
        self.__data = []


    # ----------------------------------------------------------------------------------------------------------------

    def line(self):
        # validate...
        if not self.has_regression():
            return None, None

        n = len(self)

        # init...
        sum_x = 0.0
        sum_y = 0.0

        sum_x2 = 0.0
        sum_xy = 0.0

        # sum....
        for x, y in self.__data:
            sum_x += x
            sum_y += y

            sum_x2 += x * x
            sum_xy += x * y

        # compute...
        avg_x = sum_x / n
        avg_y = sum_y / n

        d_x = (sum_x2 * n) - (sum_x * sum_x)
        d_y = (sum_xy * n) - (sum_x * sum_y)

        slope = d_y / d_x

        intercept = avg_y - (slope * avg_x)

        return slope, intercept


    def midpoint(self, ndigits=None):
        # validate...
        if not self.has_midpoint():
            return None, None

        # single value...
        if len(self) == 1:
            for timestamp, value in self.__data:
                return LocalizedDatetime.construct_from_timestamp(timestamp, self.__tzinfo), float(value)

        # multiple values...
        slope, intercept = self.line()

        # x domain...
        x_data = [x for x, _ in self.__data]

        min_x = min(x_data)
        max_x = max(x_data)

        mid_x = mean((min_x, max_x))

        rec = LocalizedDatetime.construct_from_timestamp(mid_x, self.__tzinfo)

        # y val...
        value = slope * mid_x + intercept

        return rec, value if ndigits is None else round(value, ndigits)


    # ----------------------------------------------------------------------------------------------------------------

    def min(self, ndigits=None):
        try:
            value = float(min([y for _, y in self.__data]))
            return value if ndigits is None else round(value, ndigits)

        except TypeError:
            return None

    def mid(self, ndigits=None):
        # validate...
        if not self.has_midpoint():
            return None

        try:
            # single value...
            if len(self) == 1:
                for timestamp, value in self.__data:
                    return value

            # multiple values...
            slope, intercept = self.line()

            # x domain...
            x_data = [x for x, _ in self.__data]

            min_x = min(x_data)
            max_x = max(x_data)

            mid_x = mean((min_x, max_x))

            # y val...
            value = slope * mid_x + intercept

            return value if ndigits is None else round(value, ndigits)

        except TypeError:
            return None


    def max(self, ndigits=None):
        try:
            value = float(max([y for _, y in self.__data]))
            return value if ndigits is None else round(value, ndigits)

        except TypeError:
            return None


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "LinearRegression:{tally:%s, time_relative:%s, start_timestamp:%s, tzinfo:%s, data:%s}" % \
               (self.__tally, self.__time_relative, self.__start_timestamp, self.__tzinfo, self.__data)



"""
Created on 5 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

all:
site,data_status,action_code,datetime,parameter,duration,frequency,value,unit_code,qc_code,poc,
lat,lon,GISdatum,elev,method_code,mpc,mpc_value,uncertainty,qualifiers

first time / mobile:
site,data_status,,datetime,parameter,duration,,value,unit_code,qc_code,poc,
lat1,lon1,,,,,,,

subsequent:
site,data_status,,datetime,parameter,duration,,value,unit_code,qc_code,poc,
,,,,,,,,

https://www.airnow.gov/
"""

from collections import OrderedDict

from scs_core.aqcsv.data.aqcsv_datetime import AQCSVDatetime
from scs_core.aqcsv.data.aqcsv_site import AQCSVSite

from scs_core.aqcsv.specification.method import Method
from scs_core.aqcsv.specification.mpc import MPC
from scs_core.aqcsv.specification.parameter import Parameter
from scs_core.aqcsv.specification.qc import QC
from scs_core.aqcsv.specification.unit import Unit

from scs_core.data.datum import Datum
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class AQCSVRecord(JSONable):
    """
    classdocs
    """

    STATUS_PRELIMINARY =        0
    STATUS_FINAL =              1

    ACTION_DEFAULT =            0
    ACTION_INSERT_AUTO =        1
    ACTION_UPDATE_AUTO =        2
    ACTION_INSERT_NOAUTO =      3
    ACTION_UPDATE_NOAUTO =      4
    ACTION_DELETE =             5

    GIS_DATUM =                 "WGS84"


    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def fixed_int(value, fmt):
        try:
            return fmt % int(value)
        except ValueError:
            return value


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        site_code = jdict.get('site')
        data_status = jdict.get('data_status')

        action_code = jdict.get('action_code')

        datetime_code = jdict.get('datetime')
        parameter_code = jdict.get('parameter')
        duration = jdict.get('duration')

        frequency = jdict.get('frequency')

        value = jdict.get('value')
        unit_code = jdict.get('unit')
        qc_code = jdict.get('qc')
        poc = jdict.get('poc')

        lat = jdict.get('lat')
        lon = jdict.get('lon')
        gis_datum = jdict.get('GISDatum')
        elev = jdict.get('elev')

        method_code = jdict.get('method_code')
        mpc_code = jdict.get('mpc')
        mpc_value = jdict.get('mpc_value')
        uncertainty = jdict.get('uncertainty')
        qualifiers = jdict.get('qualifiers')

        return AQCSVRecord(site_code=site_code, data_status=data_status, action_code=action_code,
                           datetime_code=datetime_code, parameter_code=parameter_code, duration=duration,
                           frequency=frequency, value=value, unit_code=unit_code, qc_code=qc_code, poc=poc,
                           lat=lat, lon=lon, gis_datum=gis_datum, elev=elev, method_code=method_code,
                           mpc_code=mpc_code, mpc_value=mpc_value, uncertainty=uncertainty, qualifiers=qualifiers)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, site_code=None, data_status=None, action_code=None,
                 datetime_code=None, parameter_code=None, duration=None,
                 frequency=None, value=None, unit_code=None, qc_code=None, poc=None,
                 lat=None, lon=None, gis_datum=None, elev=None, method_code=None,
                 mpc_code=None, mpc_value=None, uncertainty=None, qualifiers=None):
        """
        Constructor
        """
        self.__site_code = str(site_code)                           # nvarchar(12)      required
        self.__data_status = Datum.int(data_status)                 # int(1)            required

        self.__action_code = Datum.int(action_code)                 # int(1)

        self.__datetime_code = str(datetime_code)                   # nvarchar(20)      required
        self.__parameter_code = parameter_code                      # int(5) or string  required
        self.__duration = Datum.int(duration)                       # int               required

        self.__frequency = Datum.int(frequency)                     # int

        self.__value = Datum.float(value, 5)                        # numeric(10,5)     required
        self.__unit_code = Datum.int(unit_code)                     # int(3)            required
        self.__qc_code = Datum.int(qc_code)                         # int               required
        self.__poc = Datum.int(poc)                                 # int               required

        self.__lat = Datum.float(lat, 6)                            # numeric(10,6)
        self.__lon = Datum.float(lon, 6)                            # numeric(10,6)
        self.__gis_datum = gis_datum                                # nvarchar(10)
        self.__elev = Datum.int(elev)                               # int

        self.__method_code = Datum.int(method_code)                 # int(3)
        self.__mpc_code = Datum.int(mpc_code)                       # int
        self.__mpc_value = Datum.float(mpc_value, 5)                # numeric(10,5)
        self.__uncertainty = Datum.float(uncertainty, 5)            # numeric(10,5)
        self.__qualifiers = qualifiers                              # nvarchar(255)


    def __eq__(self, other):
        try:
            return \
                self.site_code == other.site_code and \
                self.data_status == other.data_status and \
                self.action_code == other.action_code and \
                self.datetime_code == other.datetime_code and \
                self.parameter_code == other.parameter_code and \
                self.duration == other.duration and \
                self.frequency == other.frequency and \
                self.value == other.value and \
                self.unit_code == other.unit_code and \
                self.qc_code == other.qc_code and \
                self.poc == other.poc and \
                self.lat == other.lat and \
                self.lon == other.lon and \
                self.gis_datum == other.gis_datum and \
                self.elev == other.elev and \
                self.method_code == other.method_code and \
                self.mpc_code == other.mpc_code and \
                self.mpc_value == other.mpc_value and \
                self.uncertainty == other.uncertainty and \
                self.qualifiers == other.qualifiers

        except AttributeError:
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['site'] = self.site_code
        jdict['data_status'] = self.data_status

        jdict['action_code'] = self.action_code

        jdict['datetime'] = self.datetime_code
        jdict['parameter'] = self.fixed_int(self.parameter_code, "%05d")
        jdict['duration'] = self.duration

        jdict['frequency'] = self.frequency

        jdict['value'] = self.value
        jdict['unit'] = self.fixed_int(self.unit_code, "%03d")
        jdict['qc'] = self.qc_code
        jdict['poc'] = self.poc

        jdict['lat'] = self.lat
        jdict['lon'] = self.lon
        jdict['GISDatum'] = self.gis_datum
        jdict['elev'] = self.elev

        jdict['method_code'] = self.fixed_int(self.method_code, "%03d")
        jdict['mpc'] = self.mpc_code
        jdict['mpc_value'] = self.mpc_value
        jdict['uncertainty'] = self.uncertainty
        jdict['qualifiers'] = self.qualifiers

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def site(self):
        return AQCSVSite.construct_from_code(self.site_code)


    def datetime(self):
        return AQCSVDatetime.construct_from_code(self.datetime_code)


    def parameter(self):
        return Parameter.instance(self.parameter_code)


    def unit(self):
        return Unit.instance(self.unit_code)


    def qc(self):
        return QC.instance(self.qc_code)


    def method(self):
        return Method.instance((self.parameter_code, self.method_code))


    def mpc(self):
        return MPC.instance(self.mpc_code)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def site_code(self):
        return self.__site_code


    @property
    def data_status(self):
        return self.__data_status


    @property
    def action_code(self):
        return self.__action_code


    @property
    def datetime_code(self):
        return self.__datetime_code


    @property
    def parameter_code(self):
        return self.__parameter_code


    @property
    def duration(self):
        return self.__duration


    @property
    def frequency(self):
        return self.__frequency


    @property
    def value(self):
        return self.__value


    @property
    def unit_code(self):
        return self.__unit_code


    @property
    def qc_code(self):
        return self.__qc_code


    @property
    def poc(self):
        return self.__poc


    @property
    def lat(self):
        return self.__lat


    @property
    def lon(self):
        return self.__lon


    @property
    def gis_datum(self):
        return self.__gis_datum


    @property
    def elev(self):
        return self.__elev


    @property
    def method_code(self):
        return self.__method_code


    @property
    def mpc_code(self):
        return self.__mpc_code


    @property
    def mpc_value(self):
        return self.__mpc_value


    @property
    def uncertainty(self):
        return self.__uncertainty


    @property
    def qualifiers(self):
        return self.__qualifiers


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AQCSVRecord:{site_code:%s, data_status:%1d, action_code:%s, datetime_code:%s, parameter_code:%s, " \
               "duration:%d, frequency:%s, value:%0.5f, unit_code:%03d, qc_code:%d, poc:%d, " \
               "lat:%s, lon:%s, gis_datum:%s, elev:%s, method_code:%s, mpc_code:%s, mpc_value:%s, " \
               "uncertainty:%s, qualifiers:%s}" % \
               (self.site_code, self.data_status, self.action_code, self.datetime_code, self.parameter_code,
                self.duration, self.frequency, self.value, self.unit_code, self.qc_code, self.poc,
                self.lat, self.lon, self.gis_datum, self.elev, self.method_code, self.mpc_code, self.mpc_value,
                self.uncertainty, self.qualifiers)


# --------------------------------------------------------------------------------------------------------------------

class AQCSVFirstRecord(AQCSVRecord):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, site_code, data_status, datetime_code, parameter_code, duration, value, unit_code,
                 qc_code, poc, lat, lon):
        """
        Constructor
        """
        super().__init__(site_code=site_code, data_status=data_status, datetime_code=datetime_code,
                         parameter_code=parameter_code, duration=duration, value=value, unit_code=unit_code,
                         qc_code=qc_code, poc=poc, lat=lat, lon=lon)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AQCSVFirstRecord:{site_code:%s, data_status:%1d, datetime_code:%s, parameter_code:%05d, " \
               "duration:%d, value:%0.5f, unit_code:%03d, qc_code:%d, poc:%d, lat:%0.6f, lon:%0.6f}" % \
               (self.site_code, self.data_status, self.datetime_code, self.parameter_code,
                self.duration, self.value, self.unit_code, self.qc_code, self.poc, self.lat, self.lon)


# --------------------------------------------------------------------------------------------------------------------

class AQCSVSubsequentRecord(AQCSVRecord):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, site_code, data_status, datetime_code, parameter_code, duration, value, unit_code,
                 qc_code, poc):
        """
        Constructor
        """
        super().__init__(site_code=site_code, data_status=data_status, datetime_code=datetime_code,
                         parameter_code=parameter_code, duration=duration, value=value, unit_code=unit_code,
                         qc_code=qc_code, poc=poc)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AQCSVSubsequentRecord:{site_code:%s, data_status:%1d, datetime_code:%s, parameter_code:%05d, " \
               "duration:%d, value:%0.5f, unit_code:%03d, qc_code:%d, poc:%d}" % \
               (self.site_code, self.data_status, self.datetime_code, self.parameter_code,
                self.duration, self.value, self.unit_code, self.qc_code, self.poc)



"""
Created on 08 Mar 2021
Modified on 21 Sep 2021

@author: Jade Page (jade.page@southcoastscience.com)

https://stackoverflow.com/questions/36780856/complete-scan-of-dynamodb-with-boto3

https://github.com/boto/botocore/issues/1688

A single manager for all required dynamoDB related functions.

"""
# --------------------------------------------------------------------------------------------------------------------
from boto3.dynamodb.conditions import Key, Attr

# --------------------------------------------------------------------------------------------------------------------


class DynamoManager(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, dynamo_client, dynamo_resource):
        self.__dynamo_client = dynamo_client
        self.__dynamo_resource = dynamo_resource

    # ----------------------------------------------------------------------------------------------------------------
    # SINGLE ITEMS
    # ----------------------------------------------------------------------------------------------------------------
    def get(self, table_name, pk, pk_val, sk=None, sk_val=None):
        # get specific item
        table = self.__dynamo_resource.Table(table_name)
        if sk and sk_val:
            key = {pk: pk_val, sk: sk_val}
        elif pk and pk_val and not sk and sk_val:
            key = {pk: pk_val}
        else:
            return None

        response = table.get_item(Key=key)
        return response['Item'] if 'Item' in response else None

    def exists(self, table_name, pk, pk_val, sk=None, sk_val=None):
        table = self.__dynamo_resource.Table(table_name)
        if sk is None and sk_val is None:
            response = table.query(
                KeyConditionExpression=Key(pk).eq(pk_val)
            )
        else:
            response = table.query(
                KeyConditionExpression=Key(pk).eq(pk_val) & Key(sk).eq(sk_val)
            )
        return response['Items'] if 'Items' in response else None

    def add(self, table_name, item):
        table = self.__dynamo_resource.Table(table_name)

        response = table.put_item(
            Item=item
        )

        return response

    def add_simple(self, table_name, pk, pk_val, sk=None, sk_val=None):
        table = self.__dynamo_resource.Table(table_name)
        if sk and sk_val:
            item = {pk: pk_val, sk: sk_val}
        elif pk and pk_val and not sk and sk_val:
            item = {pk: pk_val}
        else:
            return None

        response = table.put_item(
            Item=item
        )

        return response

    def delete(self, table_name, pk, pk_val, sk=None, sk_val=None):
        table = self.__dynamo_resource.Table(table_name)
        if pk and pk_val and sk and sk_val:
            item = {pk: pk_val, sk: sk_val}
        elif pk and pk_val and not sk and not sk_val:
            item = {pk: pk_val}
        else:
            return None

        response = table.delete_item(
            Key=item
        )

        return response

    def delete_item(self, table_name, item):
        table = self.__dynamo_resource.Table(table_name)

        response = table.delete_item(
            Key=item
        )

        return response

    def update_item(self, table_name, pk, update_expression, eav):
        table = self.__dynamo_resource.Table(table_name)
        response = table.update_item(
            Key=pk,
            UpdateExpression=update_expression,
            ExpressionAttributeValues=eav,
            ReturnValues="UPDATED_NEW"
        )

        return response

    # ----------------------------------------------------------------------------------------------------------------
    # MAIN FUNCTIONALITY
    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def build_query(table_name, pk=None, pk_val=None, sk=None, sk_val=None, keys_only=False, exact=False,
                    limit=None, fosk=False):

        kwargs = {}

        if not table_name:
            return kwargs

        # retrieve all - retrieve all batched - retrieve all keys only
        if pk_val is None and sk is None and sk_val is None:
            if keys_only:
                kwargs['ProjectionExpression'] = '#pk'
                kwargs['ExpressionAttributeNames'] = {'#pk': pk}
            if limit:
                kwargs['Limit'] = limit
            return kwargs

        # retrieve filtered on pk - retrieve filtered on pk keys only - retrieve filtered on pk batched
        if pk and pk_val and sk is None and sk_val is None and not fosk:
            if exact:
                kwargs['FilterExpression'] = (Attr(pk).eq(pk_val))
            else:
                kwargs['FilterExpression'] = (Attr(pk).contains(pk_val))
            if limit:
                kwargs['Limit'] = limit
            if keys_only:
                kwargs['ProjectionExpression'] = '#pk'
                kwargs['ExpressionAttributeNames'] = {'#pk': pk}

            return kwargs

        # retrieve double filtered - retrieve double filtered keys only - retrieve double filtered batched
        # basically the same as a get but can also be used with batch, exact and reserved keys
        if pk and pk_val and sk and sk_val and not fosk:
            if exact:
                kwargs['FilterExpression'] = Attr(pk).eq(pk_val) & Attr(sk).eq(sk_val)
            else:
                kwargs['FilterExpression'] = Attr(pk).contains(pk_val) & Attr(sk).contains(sk_val)
            if limit:
                kwargs['Limit'] = limit
            if keys_only:
                kwargs['ProjectionExpression'] = '#pk'
                kwargs['ExpressionAttributeNames'] = {'#pk': pk}

            return kwargs

        # Filter on second key - filter on second key keys only - filter on second key batched
        if pk and sk and sk_val and fosk:
            kwargs['ProjectionExpression'] = pk
            if exact:
                kwargs['FilterExpression'] = Attr(sk).eq(sk_val)
            else:
                kwargs['FilterExpression'] = Attr(sk).contains(sk_val)
            if limit:
                kwargs['Limit'] = limit
            if keys_only:
                kwargs['ProjectionExpression'] = '#pk'
                kwargs['ExpressionAttributeNames'] = {'#pk': pk}

            return kwargs


    def do_query(self, table_name, query, limited=False):
        table = self.__dynamo_resource.Table(table_name)
        to_return = []
        q = table.scan(**query)
        to_return.extend(q['Items'])
        if not limited:
            while 'LastEvaluatedKey' in q:
                query['ExclusiveStartKey'] = q['LastEvaluatedKey']
                q = table.scan(**query)
                to_return.extend(q['Items'])
        return to_return

    # ----------------------------------------------------------------------------------------------------------------
    # UTILITIES
    # ----------------------------------------------------------------------------------------------------------------

    def describe_table(self, table_name):
        response = self.__dynamo_client.describe_table(
            TableName=table_name
        )

        return response

    def return_key_schema(self, table_name):
        table = self.__dynamo_resource.Table(table_name)
        return table.key_schema


    # ----------------------------------------------------------------------------------------------------------------
    # BATCH FUNCTIONS
    # ----------------------------------------------------------------------------------------------------------------

    def batch_delete(self, table_name, items):
        table = self.__dynamo_resource.Table(table_name)
        with table.batch_writer() as batch:
            for item in items:
                batch.delete_item(item)

    def batch_write(self, table_name, items):
        table = self.__dynamo_resource.Table(table_name)
        with table.batch_writer() as batch:
            for item in items:
                batch.put_item(Item=item)


# --------------------------------------------------------------------------------------------------------------------



"""
Created on 19 Sep 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

A catalogue entry for a machine learning model, for a specific gas

document example:
{"data-set": "ref-scs-opc-116-gases-2021HA-vcal-slp-err-clean",
"period": {"start": "2021-02-10T10:15:00Z", "end": "2021-09-01T00:00:00Z"},
"primaries": {"NO2.vCal": {"path": "src.scs.env.gases.val.NO2.vCal", "min": -16.297, "avg": 9.003, "max": 67.087}},
"secondaries": {"tmp.brd": {"path": "src.scs.stat.val.tmp.brd", "min": 11.7, "avg": 26.955, "max": 47.8},
"hmd.cur": {"path": "src.scs.env.meteo.val.hmd.cur", "min": 18.7, "avg": 62.599, "max": 87.3},
"hmd.slope": {"path": "src.scs.env.meteo.val.hmd.slope", "min": -0.01, "avg": -0.0, "max": 0.015},
"tmp.cur": {"path": "src.scs.env.meteo.val.tmp.cur", "min": -1.4, "avg": 16.044, "max": 37.2},
"tmp.slope": {"path": "src.scs.env.meteo.val.tmp.slope", "min": -0.003, "avg": 0.0, "max": 0.002}},
"reference": {"path": "src.ref.NO2 (ppb)", "min": 0.0, "avg": 9.0, "max": 55.3},
"output": {"path": "exg.NO2.vE.OPCube.21HA", "min": 0.8, "avg": 9.0, "max": 50.0},
"performance": {"count": 18887, "slope": 0.91, "intercept": 0.813, "r2": 0.929, "p": 0.0, "std-err": 0.002}}
"""

import os

from collections import OrderedDict

from scs_core.data.json import JSONCatalogueEntry, JSONify
from scs_core.data.lin_regress import LinRegress
from scs_core.data.path_dict import PathDict
from scs_core.data.str import Str

from scs_core.model.catalogue.term import Term, PrimaryTerm, SecondaryTerm
from scs_core.model.catalogue.training_period import TrainingPeriod

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class ModelCompendium(JSONCatalogueEntry):
    """
    classdocs
    """

    __CATALOGUE_NAME = 'compendia'

    @classmethod
    def catalogue_location(cls):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)), cls.__CATALOGUE_NAME)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __term_path(cls, path):
        return '.'.join(path.split('.')[-2:])


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, data_set, recs, primary_values, secondary_values, output_path, outputs, reference_path,
                  references):
        period = TrainingPeriod.construct(recs)

        primaries = {cls.__term_path(path): PrimaryTerm.construct(path, values, prec=3)
                     for path, values in primary_values.items()}

        secondaries = {cls.__term_path(path): SecondaryTerm.construct(path, values, prec=3)
                       for path, values in secondary_values.items()}

        reference = Term.construct(reference_path, references, prec=1)
        output = Term.construct(output_path, outputs, prec=1)

        performance = LinRegress.construct(references, outputs, prec=3)

        return cls(data_set, period, primaries, secondaries, reference, output, performance)


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        data_set = jdict.get('data-set')
        period = TrainingPeriod.construct_from_jdict(jdict.get('period'))

        primaries = {path: PrimaryTerm.construct_from_jdict(term_jdict)
                     for path, term_jdict in jdict.get('primaries').items()}

        secondaries = {path: SecondaryTerm.construct_from_jdict(term_jdict)
                       for path, term_jdict in jdict.get('secondaries').items()}

        reference = Term.construct_from_jdict(jdict.get('reference'))
        output = Term.construct_from_jdict(jdict.get('output'))

        performance = LinRegress.construct_from_jdict(jdict.get('performance'))

        return cls(data_set, period, primaries, secondaries, reference, output, performance)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, data_set, period, primaries, secondaries, reference, output, performance):
        """
        Constructor
        """
        self.__data_set = data_set                  # string
        self.__period = period                      # TrainingPeriod

        self.__primaries = primaries                # dict of path: PrimaryTerm
        self.__secondaries = secondaries            # dict of path: SecondaryTerm
        self.__reference = reference                # Term
        self.__output = output                      # Term

        self.__performance = performance            # LinRegress

        self.__logger = Logging.getLogger()


    def __len__(self):
        return len(self.performance)


    # ----------------------------------------------------------------------------------------------------------------

    def is_in_bounds(self, datum: PathDict):
        for datum_path in datum.paths():
            term_path = self.__term_path(datum_path)

            if term_path in self.secondaries:
                if not self.secondaries[term_path].is_in_bounds(datum.node(datum_path)):
                    return False

        return True


    def preprocess(self, datum: PathDict, offset=0, include_orig=False):
        target = PathDict()

        for datum_path in datum.paths():
            term_path = self.__term_path(datum_path)
            node = datum.node(datum_path)

            if term_path in self.primaries:
                value, extr = self.primaries[term_path].preprocess(node, offset)

                if include_orig:
                    target.append(datum_path + 'Orig', node)

                target.append(datum_path, None if value is None else "%g" % value)
                target.append(datum_path + 'Extr', None if value is None else "%g" % extr)
                continue

            if term_path in self.secondaries:
                value = self.secondaries[term_path].preprocess(node)

                if include_orig:
                    target.append(datum_path + 'Orig', node)

                target.append(datum_path, None if value is None else "%g" % value)
                continue

            target.append(datum_path, node)

        self.__logger.debug("preprocess - datum: %s" % JSONify.dumps(datum))
        self.__logger.debug("preprocess - target: %s" % JSONify.dumps(target))

        return target


    def postprocess(self, primary_path, vcal_excess, model_output):
        corrected_exg = (model_output - self.performance.intercept) / self.performance.slope

        if vcal_excess > 0:
            model_vcal_max = self.primary_term(primary_path).maximum
            corrected_exg = corrected_exg * ((model_vcal_max + vcal_excess) / model_vcal_max)

        return corrected_exg


    # ----------------------------------------------------------------------------------------------------------------

    def primary_term(self, path):
        try:
            return self.primaries[self.__term_path(path)]
        except KeyError:
            return None


    def secondary_term(self, path):
        try:
            return self.secondaries[self.__term_path(path)]
        except KeyError:
            return None


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def name(self):
        return self.output.path[len('exg.'):]


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def data_set(self):
        return self.__data_set


    @property
    def period(self):
        return self.__period


    @property
    def primaries(self):
        return self.__primaries


    @property
    def secondaries(self):
        return self.__secondaries


    @property
    def reference(self):
        return self.__reference


    @property
    def output(self):
        return self.__output


    @property
    def performance(self):
        return self.__performance


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['data-set'] = self.data_set
        jdict['period'] = self.period

        jdict['primaries'] = self.primaries
        jdict['secondaries'] = self.secondaries
        jdict['reference'] = self.reference
        jdict['output'] = self.output

        jdict['performance'] = self.performance

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        primaries = Str.collection(self.primaries)
        secondaries = Str.collection(self.secondaries)

        return "ModelCompendium:{data_set:%s, data_set:%s, period:%s, primaries:%s, secondaries:%s, reference:%s, " \
               "output:%s, performance:%s}" %  \
               (self.name, self.data_set, self.period, primaries, secondaries, self.reference,
                self.output, self.performance)



"""
Created on 11 Jan 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
{"serial_number": "10-000056", "type": "810-0021-00", "calibrated_on": "YYYY-MM-DD", "dispatched_on": null,
"pt1000_v20": 1.0, "sn1": {"serial_number": "NNNNNNNNN", "sensor_type": "NOGA4", "we_electronic_zero_mv": 300,
"we_sensor_zero_mv": 6, "we_total_zero_mv": 300, "ae_electronic_zero_mv": 300, "ae_sensor_zero_mv": 1,
"ae_total_zero_mv": 300, "we_sensitivity_na_ppb": S.SSSSSSS, "we_cross_sensitivity_no2_na_ppb": -0.3,
"pcb_gain": -0.7, "we_sensitivity_mv_ppb": 0.2, "we_cross_sensitivity_no2_mv_ppb": 0.2}}
"""

import json

from scs_core.client.http_client import HTTPClient

from scs_core.gas.afe_calib import AFECalib
from scs_core.gas.sensor_calib import SensorCalib


# --------------------------------------------------------------------------------------------------------------------

class DSICalib(AFECalib):
    """
    classdocs
    """

    DSI_WRAPPER = '''
                {"serial_number": "00-000000", "type": "ISI", "calibrated_on": "YYYY-MM-DD", 
                "dispatched_on": null, "pt1000_v20": 1.0, 
                "sn1": {"serial_number": "NNNNNNNNN", "sensor_type": "A4", "we_electronic_zero_mv": 300, 
                "we_sensor_zero_mv": 0, "we_total_zero_mv": 300, "ae_electronic_zero_mv": 300, 
                "ae_sensor_zero_mv": 0, "ae_total_zero_mv": 300, "we_sensitivity_na_ppb": "S.SSSSSSS", 
                "we_cross_sensitivity_no2_na_ppb": -0.3, "pcb_gain": -0.7, "we_sensitivity_mv_ppb": 0.2, 
                "we_cross_sensitivity_no2_mv_ppb": "n/a"}}
                '''

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def download(cls, serial_number):
        http_client = HTTPClient()
        http_client.connect(AFECalib.ALPHASENSE_HOST)

        try:
            path = SensorCalib.ALPHASENSE_PATH + serial_number
            jstr = http_client.get(path, None, SensorCalib.ALPHASENSE_HEADER)

            return cls.construct_from_jdict(json.loads(jstr))

        finally:
            http_client.close()


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        sensor_calib = SensorCalib.construct_from_jdict(jdict)
        sensor_calib.set_defaults()
        sensor_calib.set_sens_mv_from_sens_na()

        return cls('DSI', sensor_calib)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, afe_type, sensor_calib):
        """
        Constructor
        """
        super().__init__(None, afe_type, None, None, None, [sensor_calib])



"""
Created on 12 Apr 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import os


# --------------------------------------------------------------------------------------------------------------------

class Filesystem(object):

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def mkdir(cls, path):
        if not path or os.path.exists(path):
            return

        head, _ = os.path.split(path)

        if head and not os.path.exists(head):
            cls.mkdir(head)

        if not os.path.exists(path):                # handles case of trailing /
            try:
                os.mkdir(path)
            except FileExistsError:
                raise FileNotFoundError(path)       # covers a Python bug


    @classmethod
    def rmdir(cls, path):
        if not path or not os.path.exists(path):
            return False

        try:
            os.rmdir(path)
            return True

        except OSError:
            return False


    @classmethod
    def rm(cls, path):
        if not path or not os.path.exists(path):
            return False

        try:
            os.remove(path)
            return True

        except OSError:
            return False


    @classmethod
    def ls(cls, path):
        if not path or not os.path.exists(path):
            return None

        return [File.construct(path, name) for name in sorted(os.listdir(path))]


# --------------------------------------------------------------------------------------------------------------------

class File(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, container, name):
        abs_filename = os.path.join(container, name)

        return File(container, name, os.path.getmtime(abs_filename), os.path.isdir(abs_filename))


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, container, name, updated, is_directory):
        """
        Constructor
        """
        self.__container = container                            # string
        self.__name = name                                      # string
        self.__updated = int(updated)                           # int timestamp
        self.__is_directory = bool(is_directory)                # bool


    # ----------------------------------------------------------------------------------------------------------------

    def path(self):
        return os.path.join(self.container, self.name)


    def has_suffix(self, suffix):
        return self.name.endswith('.' + suffix)


    def delete(self):
        return Filesystem.rm(self.path())


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def container(self):
        return self.__container


    @property
    def name(self):
        return self.__name


    @property
    def updated(self):
        return self.__updated


    @property
    def is_directory(self):
        return self.__is_directory


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "File:{container:%s, name:%s, updated:%s, is_directory:%s}" %  \
               (self.container, self.name, self.updated, self.is_directory)



"""
Created on 9 Aug 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import csv
import sys


# --------------------------------------------------------------------------------------------------------------------

class Histogram(object):
    """
    classdocs
    """

    __HEADER_BIN = ".bin"
    __HEADER_COUNT = ".count"


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, minimum, maximum, bin_count, precision, path):
        """
        Constructor
        """
        self.__minimum = minimum                                    # float
        self.__maximum = maximum                                    # float
        self.__bin_count = bin_count                                # int
        self.__precision = precision                                # int

        self.__path = path                                          # string

        self.__counts = [0] * bin_count                             # array of int
        self.__max_count = int(0)                                   # int

        self.__delta = (maximum - minimum) / bin_count              # float


    def __len__(self):
        return self.__bin_count


    # ----------------------------------------------------------------------------------------------------------------

    def append(self, datum):
        # reject out-of-range
        if datum < self.__minimum or datum > self.__maximum:
            raise ValueError("datum out of range:%f" % datum)

        # compute index...
        offset = datum - self.__minimum
        index = int(offset // self.__delta)

        # update counts...
        self.__counts[index] += 1

        if self.__counts[index] > self.__max_count:
            self.__max_count = int(self.__counts[index])

        return index, self.__counts[index]


    def to_csv(self, filename=None):
        file = sys.stdout if filename is None else open(filename, "w")
        writer = csv.writer(file)

        writer.writerow((self.__path + Histogram.__HEADER_BIN, self.__path + Histogram.__HEADER_COUNT))

        for i in range(self.bin_count):
            writer.writerow((round(self.__bin(i), self.precision), self.__counts[i]))   # format(self.__bin(i), '.6f')

        if filename is not None:
            file.close()


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def bins(self):
        return [self.__bin(i) for i in range(self.__bin_count)]


    @property
    def minimum(self):
        return self.__minimum


    @property
    def maximum(self):
        return self.__maximum


    @property
    def bin_count(self):
        return self.__bin_count


    @property
    def precision(self):
        return self.__precision


    @property
    def path(self):
        return self.__path


    @property
    def delta(self):
        return self.__delta


    @property
    def max_count(self):
        return self.__max_count


    @property
    def counts(self):
        return self.__counts


    # ----------------------------------------------------------------------------------------------------------------

    def __bin(self, index):
        return self.__minimum + (index * self.__delta)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Histogram:{minimum:%0.6f, maximum:%0.6f, bin_count:%d, bin_count:%d, delta:%0.6f, max_count:%d, " \
               "counts:%s, path:%s}" % \
                    (self.minimum, self.maximum, self.bin_count, self.precision, self.delta, self.max_count,
                     self.counts, self.path)



"""
Created on 13 Aug 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/42568262/how-to-encrypt-text-with-a-password-in-python
"""

import json
import os
import time

from abc import abstractmethod
from decimal import Decimal

from scs_core.data.datum import Datum


# --------------------------------------------------------------------------------------------------------------------

class JSONify(json.JSONEncoder):
    """
    classdocs
    """

    @classmethod
    def as_dynamo_json(cls, obj):
        if isinstance(obj, JSONable):
            return cls.as_dynamo_json(obj.as_json())

        if isinstance(obj, dict):
            return {key: cls.as_dynamo_json(value) for key, value in obj.items()}

        if isinstance(obj, list):
            return [cls.as_dynamo_json(value) for value in obj]

        if Datum.is_numeric(obj):
            return Decimal(str(obj))

        return obj


    @staticmethod
    def dumps(obj, skipkeys=False, ensure_ascii=False, check_circular=True,
              allow_nan=True, cls=None, indent=None, separators=None,
              default=None, sort_keys=False, **kw):

        handler = JSONify if cls is None else cls

        return json.dumps(obj, skipkeys=skipkeys, ensure_ascii=ensure_ascii, check_circular=check_circular,
                          allow_nan=allow_nan, cls=handler, indent=indent, separators=separators,
                          default=default, sort_keys=sort_keys, **kw)


    # ----------------------------------------------------------------------------------------------------------------

    def default(self, obj):
        if isinstance(obj, JSONable):
            return obj.as_json()

        if isinstance(obj, Decimal):
            return float(obj) if Datum.is_float(str(obj)) else int(obj)

        return json.JSONEncoder.default(self, obj)


# --------------------------------------------------------------------------------------------------------------------

class JSONable(object):
    """
    classdocs
    """

    _INDENT = 4

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def loads(cls, jstr):
        try:
            return json.loads(jstr)
        except json.decoder.JSONDecodeError:
            raise ValueError(jstr.strip())


    # ----------------------------------------------------------------------------------------------------------------

    def as_list(self, jlist):
        del jlist[:]                                    # empty the list

        for key, value in self.as_json().items():
            try:
                value = value.as_json()
            except AttributeError:
                pass

            jlist.append((key, value))                  # append the key-value pairs of the dictionary


    # ----------------------------------------------------------------------------------------------------------------

    def as_dynamo_json(self):
        return JSONify.as_dynamo_json(self)


    @abstractmethod
    def as_json(self, *args, **kwargs):
        pass


# --------------------------------------------------------------------------------------------------------------------

class JSONReport(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def load(cls, filename, skeleton=False):
        if filename is None:
            return None

        if not os.path.isfile(filename):
            return cls.construct_from_jdict(None, skeleton=skeleton)

        with open(filename, 'r') as f:
            jstr = f.read()

        return cls.construct_from_jdict(cls.loads(jstr), skeleton=skeleton)


    @classmethod
    def delete(cls, filename):
        if filename is None:
            return

        os.remove(filename)


    @classmethod
    @abstractmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        return cls()


    # ----------------------------------------------------------------------------------------------------------------

    def save(self, filename):
        if filename is None:
            return

        # data...
        jstr = JSONify.dumps(self, indent=self._INDENT)

        # file...
        tmp_filename = '.'.join((filename, str(int(time.time()))))

        try:
            with open(tmp_filename, 'w') as f:
                f.write(jstr + '\n')

        except FileNotFoundError:           # the containing directory does not exist (yet)
            return False

        # atomic operation...
        os.rename(tmp_filename, filename)

        return True


# --------------------------------------------------------------------------------------------------------------------

class JSONCatalogueEntry(JSONReport):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def list(cls):
        return [cls.__filename_to_name(item) for item in sorted(os.listdir(cls.catalogue_location()))
                if item.endswith('.json')]


    @classmethod
    def exists(cls, name):
        return name in cls.list()


    @classmethod
    def retrieve(cls, name):
        return cls.load(cls.__catalogue_entry_location(name))


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def catalogue_location(cls):
        return ''


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __catalogue_entry_location(cls, name):
        return os.path.join(cls.catalogue_location(), cls.__name_to_filename(name))


    @classmethod
    def __name_to_filename(cls, name):
        return name.replace('.', '-') + '.json'


    @classmethod
    def __filename_to_name(cls, name):
        return name.replace('-', '.')[:-len('.json')]


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def filename(self):
        return self.__catalogue_entry_location(self.name)


    # ----------------------------------------------------------------------------------------------------------------

    def store(self):
        self.save(self.filename)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    @abstractmethod
    def name(self):
        pass


# --------------------------------------------------------------------------------------------------------------------

class AbstractPersistentJSONable(JSONable):
    """
    classdocs
    """

    _SECURITY_DELAY =       3.0                                 # seconds

    __AWS_DIR =             "aws"                               # hard-coded rel path
    __CONF_DIR =            "conf"                              # hard-coded rel path
    __HUE_DIR =             "hue"                               # hard-coded rel path
    __OSIO_DIR =            "osio"                              # hard-coded rel path

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def aws_dir(cls):
        return cls.__AWS_DIR


    @classmethod
    def conf_dir(cls):
        return cls.__CONF_DIR


    @classmethod
    def hue_dir(cls):
        return cls.__HUE_DIR


    @classmethod
    def osio_dir(cls):
        return cls.__OSIO_DIR


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, last_modified=None):
        self._last_modified = last_modified                     # LocalizedDatetime


    # ----------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def save(self, manager):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def last_modified(self):
        return self._last_modified


# --------------------------------------------------------------------------------------------------------------------

class PersistentJSONable(AbstractPersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def exists(cls, manager):
        try:
            dirname, filename = cls.persistence_location()
        except NotImplementedError:
            return False

        return manager.exists(dirname, filename)


    @classmethod
    def load(cls, manager, encryption_key=None, skeleton=False):
        try:
            dirname, filename = cls.persistence_location()
        except NotImplementedError:
            return None

        if not manager.exists(dirname, filename):
            return cls.construct_from_jdict(None, skeleton=skeleton)

        try:
            jstr, last_modified = manager.load(dirname, filename, encryption_key=encryption_key)
        except (KeyError, ValueError) as ex:            # caused by incorrect encryption_key
            time.sleep(cls._SECURITY_DELAY)
            raise ex

        try:
            obj = cls.construct_from_jdict(cls.loads(jstr), skeleton=skeleton)
            obj._last_modified = last_modified
            return obj

        except (AttributeError, TypeError):
            return None


    @classmethod
    def delete(cls, manager):
        try:
            dirname, filename = cls.persistence_location()
            manager.remove(dirname, filename)

        except NotImplementedError:
            pass


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        return PersistentJSONable()


    @classmethod
    @abstractmethod
    def persistence_location(cls):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    def save(self, manager, encryption_key=None):
        self._last_modified = None                          # last_modified field shall be restored by load(..)

        dirname, filename = self.persistence_location()
        jstr = JSONify.dumps(self, indent=self._INDENT)

        manager.save(jstr, dirname, filename, encryption_key=encryption_key)


# --------------------------------------------------------------------------------------------------------------------

class MultiPersistentJSONable(AbstractPersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def list(cls, manager):
        try:
            dirname, filename = cls.persistence_location(None)
        except NotImplementedError:
            return None

        suffix_len = len(filename) + 1
        items = manager.list(manager.scs_path(), dirname)

        return [item[:-suffix_len] for item in items if item.endswith(filename)]


    @classmethod
    def exists(cls, manager, name=None):
        try:
            dirname, filename = cls.persistence_location(name)
        except NotImplementedError:
            return False

        return manager.exists(dirname, filename)


    @classmethod
    def load(cls, manager, name=None, encryption_key=None, skeleton=False):
        try:
            dirname, filename = cls.persistence_location(name)
        except NotImplementedError:
            return None

        if not manager.exists(dirname, filename):
            return cls.construct_from_jdict(None, name=name, skeleton=skeleton)

        try:
            jstr, last_modified = manager.load(dirname, filename, encryption_key=encryption_key)

        except (KeyError, ValueError) as ex:            # caused by incorrect encryption_key
            time.sleep(cls._SECURITY_DELAY)
            raise ex

        try:
            obj = cls.construct_from_jdict(cls.loads(jstr), name=name, skeleton=skeleton)
            obj._last_modified = last_modified
            return obj

        except (AttributeError, TypeError):
            return None


    @classmethod
    def delete(cls, manager, name=None):
        try:
            dirname, filename = cls.persistence_location(name)
            manager.remove(dirname, filename)

        except NotImplementedError:
            pass


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def construct_from_jdict(cls, jdict, name=None, skeleton=False):
        return PersistentJSONable()


    @classmethod
    @abstractmethod
    def persistence_location(cls, name):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, name, last_modified=None):
        super().__init__(last_modified=last_modified)

        self.__name = name                                  # string


    # ----------------------------------------------------------------------------------------------------------------

    def save(self, manager, encryption_key=None):
        self._last_modified = None                          # last_modified field shall be restored by load(..)

        dirname, filename = self.persistence_location(self.name)
        jstr = JSONify.dumps(self, indent=self._INDENT)

        manager.save(jstr, dirname, filename, encryption_key=encryption_key)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def name(self):
        return self.__name


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MultiPersistentJSONable:{name:%s}" % self.name



"""
Created on 21 Mar 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

{
  "id": "southcoastscience-dev",
  "name": "South Coast Science - Dev",
  "email": "opensensors-dev@southcoastscience.com",
  "month": "2016-11"
}
"""

from collections import OrderedDict

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class User(JSONable):
    """
    classdocs
   """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        id = jdict.get('id')
        name = jdict.get('name')
        email = jdict.get('email')
        
        start = jdict.get('month')

        return User(id, name, email, None, start)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, id, name, email, password, start):
        """
        Constructor
        """
        self.__id = id                  # string
        self.__name = name              # string
        self.__email = email            # string

        self.__password = password      # string

        self.__start = start            # string (year-month)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        if self.id is not None:
            jdict['id'] = self.id

        if self.name is not None:
            jdict['name'] = self.name

        if self.email is not None:
            jdict['email'] = self.email

        if self.password is not None:
            jdict['password'] = self.password

        if self.start is not None:
            jdict['month'] = self.start

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def start_year(self):
        if self.start is None:
            return None

        year = self.start.split('-')[0]

        return int(year)


    def start_month(self):
        if self.start is None:
            return None

        month = self.start.split('-')[1]

        return int(month)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def id(self):
        return self.__id


    @property
    def name(self):
        return self.__name


    @property
    def email(self):
        return self.__email


    @property
    def password(self):
        return self.__password


    @property
    def start(self):
        return self.__start


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "User:{id:%s, name:%s, email:%s, password:%s, start:%s}" % \
               (self.id, self.name, self.email, self.password, self.start)



"""
Created on 13 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import socket

import urllib.parse

from collections import OrderedDict

from scs_core.osio.client.rest_client import RESTClient

from scs_core.osio.data.device_topic import DeviceTopic
from scs_core.osio.data.topic_summary import TopicSummary
from scs_core.osio.data.topic_metadata import TopicMetadata
from scs_core.osio.data.user_topic import UserTopic

from scs_core.osio.manager.message_manager import NextMessageQuery


# --------------------------------------------------------------------------------------------------------------------

class TopicManager(object):
    """
    classdocs
    """
    __FINDER_BATCH_SIZE = 100

    __TIMEOUT = 10.0                            # Required because of OSIO non-response on CRUD operations.


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, api_key):
        """
        Constructor
        """
        self.__rest_client = RESTClient(api_key)


    # ----------------------------------------------------------------------------------------------------------------

    def find(self, topic_path):
        if topic_path is None:
            return None

        request_path = '/v1/topics/' + urllib.parse.quote(topic_path, '')

        # request...
        self.__rest_client.connect()

        try:
            response_jdict = self.__rest_client.get(request_path)
        except RuntimeError:
            response_jdict = None

        self.__rest_client.close()

        topic = TopicMetadata.construct_from_jdict(response_jdict)

        return topic


    def find_for_org(self, org_id, partial_topic_path=None, topic_schema=None):
        topics = []

        # request...
        self.__rest_client.connect()

        try:
            for batch in self.__find(org_id, partial_topic_path, topic_schema):
                topics.extend(batch)

        finally:
            self.__rest_client.close()

        return topics


    def find_for_user(self, user_id):
        topics = []

        # request...
        self.__rest_client.connect()

        try:
            for batch in self.__find_for_user(user_id):
                topics.extend(batch)

        finally:
            self.__rest_client.close()

        return topics


    def find_for_device(self, client_id, start_date, end_date):
        request_path = '/v1/messages/device/' + client_id
        params = {'start-date': start_date.as_iso8601(), 'end-date': end_date.as_iso8601()}

        topics = {}

        # request...
        self.__rest_client.connect()

        try:
            while True:
                jdict = self.__rest_client.get(request_path, params=params)

                # messages...
                for message in jdict['messages']:
                    if message['topic'] not in topics:
                        topics[message['topic']] = DeviceTopic.construct_from_message_jdict(message)

                # next...
                next_query = NextMessageQuery.construct_from_uri(jdict.get('next'))

                if next_query is None:
                    break

                params['start-date'] = next_query.start_date.as_iso8601()
                params['end-date'] = next_query.end_date.as_iso8601()

            sorted_topics = OrderedDict(sorted(topics.items()))

            return list(sorted_topics.values())

        finally:
            self.__rest_client.close()


    # ----------------------------------------------------------------------------------------------------------------

    def __find(self, org_id, partial_topic_path=None, topic_schema=None):
        request_path = '/v2/orgs/' + org_id + '/topics'
        params = {'offset': 0, 'count': self.__FINDER_BATCH_SIZE}

        while True:
            topics = []

            # request...
            response_jdict = self.__rest_client.get(request_path, params=params)

            if response_jdict is None:
                return

            # topics...
            topics_jdict = response_jdict.get('topics')

            if topics_jdict:
                for topic_jdict in topics_jdict:
                    topic = TopicSummary.construct_from_jdict(topic_jdict)

                    if partial_topic_path is not None and partial_topic_path not in topic.path:
                        continue

                    if topic_schema is not None and (topic.schema_id != topic_schema):
                        continue

                    topics.append(topic)

            yield topics

            if len(topics_jdict) == 0:
                return

            # next...
            params['offset'] += len(topics_jdict)


    def __find_for_user(self, user_id):
        request_path = '/v1/users/' + user_id + '/topics'
        params = {'offset': 0, 'count': self.__FINDER_BATCH_SIZE}

        while True:
            # request...
            topics_jdict = self.__rest_client.get(request_path, params=params)

            # topics...
            topics = [UserTopic.construct_from_jdict(topic_jdict) for topic_jdict in topics_jdict]

            yield topics

            if len(topics_jdict) == 0:
                return

            # next...
            params['offset'] += len(topics_jdict)


    # ----------------------------------------------------------------------------------------------------------------

    def create(self, topic):
        request_path = '/v2/topics'

        # request...
        self.__rest_client.connect(timeout=self.__TIMEOUT)

        try:
            response = self.__rest_client.post(request_path, topic.as_json())

        except socket.timeout:
            return True

        finally:
            self.__rest_client.close()

        success = response == topic.path

        return success


    def update(self, topic_path, topic):
        request_path = '/v1/topics/' + topic_path

        # request...
        self.__rest_client.connect(timeout=self.__TIMEOUT)

        try:
            self.__rest_client.put(request_path, topic.as_json())

        except socket.timeout:
            pass

        finally:
            self.__rest_client.close()


    def delete(self, topic_path):
        request_path = '/v1/topics/' + urllib.parse.quote(topic_path, '')

        # request...
        self.__rest_client.connect(timeout=self.__TIMEOUT)

        try:
            response = self.__rest_client.delete(request_path)

        except socket.timeout:
            return True

        finally:
            self.__rest_client.close()

        success = response == ''

        return success


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "TopicManager:{rest_client:%s}" % self.__rest_client



"""
Created on 24 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from collections import OrderedDict

from scs_core.data.datum import Datum
from scs_core.gas.sensor_calib import SensorCalib


# --------------------------------------------------------------------------------------------------------------------

class PIDCalib(SensorCalib):
    """
    classdocs
    """

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        serial_number = jdict.get('serial_number')
        sensor_type = jdict.get('sensor_type')

        pid_elc_mv = jdict.get('pid_zero_mv')
        pid_sens_mv = jdict.get('pid_sensitivity_mv_ppm')

        return cls(serial_number, sensor_type, pid_elc_mv, pid_sens_mv)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, serial_number, sensor_type, pid_elc_mv, pid_sens_mv):
        """
        Constructor
        """
        SensorCalib.__init__(self, serial_number, sensor_type)

        self.__pid_elc_mv = Datum.int(pid_elc_mv)                 # PID electronic zero                   mV
        self.__pid_sens_mv = Datum.float(pid_sens_mv, 6)          # PID sensitivity                       mV / ppm


    def __eq__(self, other):
        try:
            return self.serial_number == other.serial_number and self.sensor_type == other.sensor_type and \
                   self.pid_elc_mv == other.pid_elc_mv and self.pid_sens_mv == other.pid_sens_mv

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def set_defaults(self):
        pass                                # PID may have null values


    def set_sens_mv_from_sens_na(self):
        pass                                # sensitivity nA data is not available


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['serial_number'] = self.serial_number
        jdict['sensor_type'] = self.sensor_type

        jdict['pid_zero_mv'] = self.pid_elc_mv
        jdict['pid_sensitivity_mv_ppm'] = self.pid_sens_mv

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def pid_elc_mv(self):
        return self.__pid_elc_mv


    @property
    def pid_sens_mv(self):
        return self.__pid_sens_mv


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "PIDCalib:{serial_number:%s, sensor_type:%s, pid_elc_mv:%s, pid_sens_mv:%s}" % \
                    (self.serial_number, self.sensor_type, self.pid_elc_mv, self.pid_sens_mv)



"""
Created on 19 Feb 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from scs_core.osio.data.device import Device


# --------------------------------------------------------------------------------------------------------------------

class ProjectClient(object):
    """
    classdocs
    """

    DEVICE_DESCRIPTION =       "South Coast Science air quality monitoring client"


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def tags(cls):
        return ['SCS']


    @classmethod
    def create(cls, name, api_auth, description, tags):
        client_id = None
        desc = cls.DEVICE_DESCRIPTION if description is None else description
        password = None
        password_is_locked = None
        device_type = 'client'
        batch = None
        org_id = api_auth.org_id
        owner_id = None

        device = Device(client_id, name, desc, password, password_is_locked, None,
                        device_type, batch, org_id, owner_id, tags)

        return device


    @classmethod
    def update(cls, existing, description, tags=None):
        client_id = None
        name = existing.name
        password = None
        password_is_locked = None
        device_type = existing.device_type
        batch = existing.batch
        org_id = None
        owner_id = None

        device_tags = existing.tags if tags is None else tags

        if description:
            desc = description
        else:
            desc = existing.description

        device = Device(client_id, name, desc, password, password_is_locked, None,
                        device_type, batch, org_id, owner_id, device_tags)

        return device



"""
Created on 4 Aug 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/43717757/commas-and-double-quotes-in-csv-files
"""

import csv
import sys

from scs_core.csv.csv_dict import CSVHeader

from scs_core.data.json import JSONify
from scs_core.data.str import Str


# --------------------------------------------------------------------------------------------------------------------

class CSVReader(object):
    """
    classdocs
    """

    __REPRESENTATIONS_OF_NULL = ('', 'NULL')

    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def list_dialects():
        return csv.list_dialects()


    @staticmethod
    def __recast(value):
        if value is None:
            return None

        try:
            return int(value)
        except ValueError:
            pass

        try:
            return float(value)
        except ValueError:
            pass

        if value.upper() == 'TRUE':
            return True

        if value.upper() == 'FALSE':
            return False

        return value


    @classmethod
    def __renullify(cls, value):
        try:
            return None if value.upper() in cls.__REPRESENTATIONS_OF_NULL else value
        except AttributeError:
            return value


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_for_file(cls, filename, cast=True, nullify=False, start_row=0):
        iterable = sys.stdin if filename is None else open(filename, "r")

        return cls(iterable, filename=filename, cast=cast, nullify=nullify,
                   start_row=start_row)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, iterable, filename=None, cast=True, nullify=False, start_row=0):
        """
        Constructor
        """
        self.__iterable = iterable                                  # iterable
        self.__filename = filename                                  # string
        self.__cast = bool(cast)                                    # bool
        self.__nullify = bool(nullify)                              # bool
        self.__start_row = int(start_row)                           # int

        try:
            self.__reader = csv.reader(iterable, quotechar='"', delimiter=',', quoting=csv.QUOTE_ALL,
                                       skipinitialspace=True)

            try:
                paths = next(self.__reader)
            except StopIteration:                                       # no input
                paths = []

            self.__read_count = 0                                       # int

            self.__header = CSVHeader.construct_from_paths(paths)       # CSVHeader

        except csv.Error as ex:
            raise CSVReaderException(ex)


    # ----------------------------------------------------------------------------------------------------------------

    def close(self):
        if self.__filename is None:
            return

        self.__iterable.close()


    # ----------------------------------------------------------------------------------------------------------------

    def rows(self):
        row_number = -1

        try:
            for row in self.__reader:
                if len(row) == 0:
                    continue

                row_number += 1

                if row_number < self.__start_row:
                    continue

                if self.__nullify:
                    row = [self.__renullify(cell) for cell in row]

                if self.__cast:
                    row = [self.__recast(cell) for cell in row]

                datum = self.__header.as_dict([cell for cell in row])

                yield JSONify.dumps(datum)

                self.__read_count += 1

        except csv.Error as ex:
            raise CSVReaderException(ex)            # typically on the last line of a badly-closed CSV file


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def filename(self):
        return self.__filename


    @property
    def read_count(self):
        return self.__read_count


    @property
    def header(self):
        return self.__header


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        iterable = self.__iterable.__class__.__name__

        return "CSVReader:{iterable:%s, filename:%s, cast:%s, nullify:%s, " \
               "start_row:%s, read_count:%s, header:%s}" % \
               (iterable, self.filename, self.__cast, self.__nullify,
                self.__start_row, self.read_count, Str.collection(list(self.header.paths())))


# --------------------------------------------------------------------------------------------------------------------

class CSVReaderException(RuntimeError):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, *args):
        super().__init__(*args)



"""
Created on 26 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
4-Way AFE:
{"serial_number":"250011","type":"810-00000","pt1000_v20":2.0,"calibrated_on":null,"dispatched_on":null,
"sn1":{"serial_number":"123456789","sensor_type":"IRMA1","ae_total_zero_mv":"287.0","we_total_zero_mv":"284.0", ...},
"sn2":{"serial_number":"123456789","sensor_type":"IRMA1","ae_total_zero_mv":"287.0","we_total_zero_mv":"284.0", ...},
"sn3":{"serial_number":"123456789","sensor_type":"IRMA1","ae_total_zero_mv":"280.0","we_total_zero_mv":"284.0", ...},
"sn4":{"serial_number":"123456789","sensor_type":"IRMA1","ae_total_zero_mv":"313.0","we_total_zero_mv":"305.0", ...}}
"""

import json

from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum
from scs_core.data.json import PersistentJSONable
from scs_core.data.str import Str
from scs_core.data.timedelta import Timedelta

from scs_core.client.http_client import HTTPClient

from scs_core.gas.afe.pt1000_calib import Pt1000Calib

from scs_core.gas.sensor import Sensor
from scs_core.gas.sensor_calib import SensorCalib


# --------------------------------------------------------------------------------------------------------------------

class CalibCurrency(object):
    """
    classdocs
    """

    __TIME_OFFSET = Timedelta(hours=12)

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def age(cls, calibrated_on, rec):
        calibrated = LocalizedDatetime.construct_from_date(calibrated_on)
        calibrated_noon = calibrated + cls.__TIME_OFFSET

        age = rec - calibrated_noon

        return int(age.total_seconds())


# --------------------------------------------------------------------------------------------------------------------

class AFECalib(PersistentJSONable):
    """
    classdocs
    """

    ALPHASENSE_HOST =       "www.alphasense-technology.co.uk"
    ALPHASENSE_PATH =       "/api/v1/boards/"
    ALPHASENSE_HEADER =     {"Accept": "application/json"}

    TEST_LOAD = '''
                {"serial_number": "1", "type": "test-load", "calibrated_on": "2020-01-01", "dispatched_on": null, 
                "pt1000_v20": 1.0, 

                "sn1": {"serial_number": "01", "sensor_type": "SN1", "we_electronic_zero_mv": 1, 
                "we_sensor_zero_mv": 1, "we_total_zero_mv": 1, "ae_electronic_zero_mv": 1, "ae_sensor_zero_mv": 1, 
                "ae_total_zero_mv": 1, "we_sensitivity_na_ppb": 1.0, "we_cross_sensitivity_no2_na_ppb": "n/a", 
                "pcb_gain": 1.0, "we_sensitivity_mv_ppb": 1.0, "we_cross_sensitivity_no2_mv_ppb": "n/a"}, 

                "sn2": {"serial_number": "02", "sensor_type": "SN2", "we_electronic_zero_mv": 1, 
                "we_sensor_zero_mv": 1, "we_total_zero_mv": 1, "ae_electronic_zero_mv": 1, "ae_sensor_zero_mv": 1, 
                "ae_total_zero_mv": 1, "we_sensitivity_na_ppb": 1.0, "we_cross_sensitivity_no2_na_ppb": "n/a", 
                "pcb_gain": 1.0, "we_sensitivity_mv_ppb": 1.0, "we_cross_sensitivity_no2_mv_ppb": "n/a"}, 

                "sn3": {"serial_number": "03", "sensor_type": "SN3", "we_electronic_zero_mv": 1, 
                "we_sensor_zero_mv": 1, "we_total_zero_mv": 1, "ae_electronic_zero_mv": 1, "ae_sensor_zero_mv": 1, 
                "ae_total_zero_mv": 1, "we_sensitivity_na_ppb": 1.0, "we_cross_sensitivity_no2_na_ppb": "n/a", 
                "pcb_gain": 1.0, "we_sensitivity_mv_ppb": 1.0, "we_cross_sensitivity_no2_mv_ppb": "n/a"}, 

                "sn4": {"serial_number": "04", "sensor_type": "SN4", "we_electronic_zero_mv": 1, 
                "we_sensor_zero_mv": 1, "we_total_zero_mv": 1, "ae_electronic_zero_mv": 1, "ae_sensor_zero_mv": 1, 
                "ae_total_zero_mv": 1, "we_sensitivity_na_ppb": 1.0, "we_cross_sensitivity_no2_na_ppb": "n/a", 
                "pcb_gain": 1.0, "we_sensitivity_mv_ppb": 1.0, "we_cross_sensitivity_no2_mv_ppb": "n/a"}}    
                '''


    # ----------------------------------------------------------------------------------------------------------------

    __FILENAME =    "afe_calib.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def download(cls, serial_number):
        http_client = HTTPClient()
        http_client.connect(AFECalib.ALPHASENSE_HOST)

        try:
            path = AFECalib.ALPHASENSE_PATH + serial_number
            jdict = json.loads(http_client.get(path, None, AFECalib.ALPHASENSE_HEADER))

            return cls.construct_from_jdict(jdict)

        finally:
            http_client.close()


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        serial_number = jdict.get('serial_number')
        afe_type = jdict.get('type')

        calibrated_on = Datum.date(jdict.get('calibrated_on'))
        dispatched_on = Datum.date(jdict.get('dispatched_on'))

        pt1000_v20 = jdict.get('pt1000_v20')
        pt1000_calib = None if pt1000_v20 is None else Pt1000Calib(calibrated_on, pt1000_v20)

        sensor_calibs = []

        for key in sorted(jdict.keys()):
            if key[:2] == "sn":
                if jdict[key] is None:
                    sensor_calibs.append(None)
                    continue

                sensor_calibs.append(SensorCalib.construct_from_jdict(jdict[key]))

        return cls(serial_number, afe_type, calibrated_on, dispatched_on, pt1000_calib, sensor_calibs)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, serial_number, afe_type, calibrated_on, dispatched_on, pt1000_calib, sensor_calibs):
        """
        Constructor
        """
        super().__init__()

        self.__serial_number = serial_number            # string
        self.__afe_type = afe_type                      # string

        self.__calibrated_on = calibrated_on            # date
        self.__dispatched_on = dispatched_on            # date

        self.__pt1000_calib = pt1000_calib              # Pt1000Calib

        self.__sensor_calibs = sensor_calibs            # array of SensorCalib


    def __eq__(self, other):
        try:
            if len(self) != len(other):
                return False

            for i in range(len(self)):
                if self.sensor_calib(i) != other.sensor_calib(i):
                    return False

            return self.serial_number == other.serial_number and self.afe_type == other.afe_type and \
                self.calibrated_on == other.calibrated_on and self.dispatched_on == other.dispatched_on and \
                self.pt1000_calib == other.pt1000_calib

        except (TypeError, AttributeError):
            return False


    def __len__(self):
        return len(self.__sensor_calibs)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['serial_number'] = self.serial_number
        jdict['type'] = self.afe_type

        jdict['calibrated_on'] = self.calibrated_on.isoformat() if self.calibrated_on else None
        jdict['dispatched_on'] = self.dispatched_on.isoformat() if self.dispatched_on else None

        jdict['pt1000_v20'] = self.pt1000_calib.v20 if self.pt1000_calib else None

        for i in range(len(self.__sensor_calibs)):
            jdict['sn' + str(i + 1)] = self.__sensor_calibs[i]

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def sensors(self, afe_baseline):
        sensors = []
        for i in range(len(self)):
            calib = self.sensor_calib(i)
            sensor = None if calib is None else calib.sensor(afe_baseline.sensor_baseline(i))

            sensors.append(sensor)

        return sensors


    def sensor_calibs(self):           # returns dict of gas_name: SensorCalib
        calibs = {}

        for sensor_calib in self.__sensor_calibs:
            if not sensor_calib:
                continue

            sensor = Sensor.find(sensor_calib.serial_number)
            name = sensor.gas_name

            if name in calibs:
                raise ValueError("duplicate gas name: %s" % name)

            calibs[sensor.gas_name] = sensor_calib

        return calibs


    def gas_names(self):
        names = []
        for sensor_calib in self.__sensor_calibs:
            if sensor_calib is None:
                continue

            sensor = Sensor.find(sensor_calib.serial_number)
            name = sensor.gas_name

            names.append(name)

        return names


    def has_unique_gas_names(self):
        names = set()

        for sensor_calib in self.__sensor_calibs:
            if sensor_calib is None:
                continue

            sensor = Sensor.find(sensor_calib.serial_number)
            name = sensor.gas_name

            if name in names:
                return False

            names.add(name)

        return True


    def sensor_index(self, gas_name):
        for i in range(len(self.__sensor_calibs)):
            sensor_calib = self.__sensor_calibs[i]

            if sensor_calib is None:
                continue

            sensor = Sensor.find(sensor_calib.serial_number)

            if sensor.gas_name == gas_name:
                return i

        return None


    # ----------------------------------------------------------------------------------------------------------------

    def age(self):
        return self.age_at(LocalizedDatetime.now())


    def age_at(self, rec):
        return CalibCurrency.age(self.calibrated_on, rec)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def serial_number(self):
        return self.__serial_number


    @property
    def afe_type(self):
        return self.__afe_type


    @property
    def calibrated_on(self):
        return self.__calibrated_on


    @calibrated_on.setter
    def calibrated_on(self, calibrated_on):
        self.__calibrated_on = calibrated_on


    @property
    def dispatched_on(self):
        return self.__dispatched_on


    @property
    def pt1000_calib(self):
        return self.__pt1000_calib


    def sensor_calib(self, i):
        return self.__sensor_calibs[i]


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        cls = self.__class__.__name__
        return cls + ":{serial_number:%s, afe_type:%s, calibrated_on:%s, " \
                     "dispatched_on:%s, pt1000_calib:%s, sensor_calibs:%s}" %  \
                     (self.serial_number, self.afe_type, self.calibrated_on,
                      self.dispatched_on, self.pt1000_calib, Str.collection(self.__sensor_calibs))



"""
Created on 11 Aug 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

The TimezoneConf subsystem enables the device user to set the reported timezone independently from the
system timezone. If no timezone_conf.json document is found, then the system "localzone" is used.

https://stackoverflow.com/questions/13866926/python-pytz-list-of-timezones

example JSON:
{"set-on": "2017-08-12T11:20:28.740+00:00", "name": "Europe/London"}
"""

from collections import OrderedDict

from tzlocal import get_localzone

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import PersistentJSONable

from scs_core.location.timezone import Timezone


# --------------------------------------------------------------------------------------------------------------------

class TimezoneConf(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "timezone_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def system_name(cls):
        return get_localzone().zone


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls(None, None) if skeleton else None

        set_on = LocalizedDatetime.construct_from_iso8601(jdict.get('set-on'))
        name = jdict.get('name')

        return cls(set_on, name)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, set_on, name):
        """
        Constructor
        """
        super().__init__()

        self.__set_on = set_on                          # LocalizedDatetime
        self.__name = name                              # a Pytz timezone name or None


    def __eq__(self, other):
        try:
            return self.set_on == other.set_on and self.name == other.name

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def save(self, manager, encryption_key=None):
        if self.__set_on is None:
            self.__set_on = LocalizedDatetime.now()

        super().save(manager, encryption_key=encryption_key)


    # ----------------------------------------------------------------------------------------------------------------

    def timezone(self):
        return Timezone(self.reporting_name())


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['set-on'] = None if self.set_on is None else self.set_on.as_iso8601()
        jdict['name'] = self.name

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def uses_system_name(self):
        return self.name is None


    def reporting_name(self):
        return self.system_name() if self.uses_system_name() else self.name


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def set_on(self):
        return self.__set_on


    @property
    def name(self):
        return self.__name


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "TimezoneConf:{set_on:%s, name:%s}" % (self.set_on, self.name)



"""
Created on 21 Sep 2020

@author: Jade Page (jade.page@southcoastscience.com)

DESCRIPTION The aws_group is a class used by aws_group_setup to collect information from the amazon greengrass
api
"""

import sys

from collections import OrderedDict

from scs_core.data.json import JSONable
from scs_core.data.path_dict import PathDict


# --------------------------------------------------------------------------------------------------------------------

class AWSGroup(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, group_name, client):
        """
        Constructor
        """
        self.__client = client
        self.__group_info = PathDict()
        self.__group_info.append("GroupName", [group_name])
        self.__verbose_group_info = PathDict()


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['info'] = self.__group_info
        jdict['v-info'] = self.__verbose_group_info

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def get_group_info_from_name(self):
        groups = []

        res = self.__client.list_groups(MaxResults='1000')
        groups.extend(res["Groups"])
        while "NextToken" in res:
            res = self.__client.list_groups(MaxResults='1000', NextToken=res["NextToken"])
            groups.extend(res["Groups"])

        # print("response...", file=sys.stderr)
        # print(response, file=sys.stderr)

        for group in groups:
            this_name = group["Name"]

            if this_name != self.__group_info.node("GroupName")[0]:
                continue

            self.__group_info.append("GroupID", group["Id"])
            self.__group_info.append("GroupLatestVersionID", group["LatestVersion"])
            self.__group_info.append("LastUpdated", group["LastUpdatedTimestamp"])
            return self.__group_info

        return None


    # ----------------------------------------------------------------------------------------------------------------

    def get_group_arns(self):
        response = self.__client.get_group_version(
            GroupId=self.__group_info.node("GroupID"),
            GroupVersionId=self.__group_info.node("GroupLatestVersionID")
        )
        v_group = PathDict(response)
        sub_v_group = v_group.node("Definition")

        if "CoreDefinitionVersionArn" in sub_v_group:
            self.__group_info.append("CoreDefinitionVersionArn",
                                     sub_v_group["CoreDefinitionVersionArn"])

        if "FunctionDefinitionVersionArn" in sub_v_group:
            self.__group_info.append("FunctionDefinitionVersionArn",
                                     sub_v_group["FunctionDefinitionVersionArn"])

        if "LoggerDefinitionVersionArn" in sub_v_group:
            self.__group_info.append("LoggerDefinitionVersionArn",
                                     sub_v_group["LoggerDefinitionVersionArn"])

        if "ResourceDefinitionVersionArn" in sub_v_group:
            self.__group_info.append("ResourceDefinitionVersionArn",
                                     sub_v_group["ResourceDefinitionVersionArn"])

        if "SubscriptionDefinitionVersionArn" in sub_v_group:
            self.__group_info.append("SubscriptionDefinitionVersionArn",
                                     sub_v_group["SubscriptionDefinitionVersionArn"])


    # ----------------------------------------------------------------------------------------------------------------

    def output_current_info(self):
        if self.__group_info.has_path("CoreDefinitionVersionArn"):
            self.get_core_definition()
        if self.__group_info.has_path("LoggerDefinitionVersionArn"):
            self.get_logger_definition()
        if self.__group_info.has_path("ResourceDefinitionVersionArn"):
            self.get_resource_definition()
        if self.__group_info.has_path("FunctionDefinitionVersionArn"):
            self.get_function_definition()
        if self.__group_info.has_path("SubscriptionDefinitionVersionArn"):
            self.get_subscription_definition_version()


    # ----------------------------------------------------------------------------------------------------------------

    def get_core_definition(self):

        if not self.__group_info.has_path("CoreDefinitionVersionArn"):
            return
        if not self.__group_info.node("CoreDefinitionVersionArn"):
            return

        arn = self.__split_id_from_version_id("core", self.__group_info.node("CoreDefinitionVersionArn"))
        response = self.__client.get_core_definition_version(
            CoreDefinitionId=arn[0],
            CoreDefinitionVersionId=arn[1]
        )
        self.__verbose_group_info.append("Core Definition Response", response)


    # ----------------------------------------------------------------------------------------------------------------

    def get_function_definition(self):
        if not self.__group_info.has_path("FunctionDefinitionVersionArn"):
            print("Group missing function definition version", file=sys.stderr)
            return

        if not self.__group_info.node("FunctionDefinitionVersionArn"):
            print("Group missing function definition version", file=sys.stderr)
            return

        arn = self.__split_id_from_version_id("function", self.__group_info.node("FunctionDefinitionVersionArn"))
        response = self.__client.get_function_definition_version(
            FunctionDefinitionId=arn[0],
            FunctionDefinitionVersionId=arn[1]
        )

        self.__verbose_group_info.append("Function Definition Response", response)


    # ----------------------------------------------------------------------------------------------------------------

    def get_logger_definition(self):

        if not self.__group_info.has_path("LoggerDefinitionVersionArn"):
            return
        if not self.__group_info.node("LoggerDefinitionVersionArn"):
            return

        arn = self.__split_id_from_version_id("logger", self.__group_info.node("LoggerDefinitionVersionArn"))
        response = self.__client.get_logger_definition_version(
            LoggerDefinitionId=arn[0],
            LoggerDefinitionVersionId=arn[1]
        )
        self.__verbose_group_info.append("Logger Definition Response", response)


    # ----------------------------------------------------------------------------------------------------------------

    def get_resource_definition(self):

        if not self.__group_info.has_path("ResourceDefinitionVersionArn"):
            return
        if not self.__group_info.node("ResourceDefinitionVersionArn"):
            return

        arn = self.__split_id_from_version_id("resources", self.__group_info.node("ResourceDefinitionVersionArn"))
        response = self.__client.get_resource_definition_version(
            ResourceDefinitionId=arn[0],
            ResourceDefinitionVersionId=arn[1]
        )
        self.__verbose_group_info.append("Resource Definition Response", response)


    # ----------------------------------------------------------------------------------------------------------------

    def get_subscription_definition_version(self):

        if not self.__group_info.has_path("SubscriptionDefinitionVersionArn"):
            return
        if not self.__group_info.node("SubscriptionDefinitionVersionArn"):
            return

        arn = self.__split_id_from_version_id("subscriptions",
                                              self.__group_info.node("SubscriptionDefinitionVersionArn"))

        response = self.__client.get_subscription_definition_version(
            SubscriptionDefinitionId=arn[0],
            SubscriptionDefinitionVersionId=arn[1]
        )
        self.__verbose_group_info.append("Subscription Definition Response", response)


    # ----------------------------------------------------------------------------------------------------------------

    def __append_to_group_info(self, path, value):
        self.__group_info.append(path, value)


    # ----------------------------------------------------------------------------------------------------------------

    def retrieve_node(self, path):
        if self.__group_info.has_path(path):
            return self.__group_info.node(path)
        else:
            return "-"

    def return_verbose_info(self):
        return self.__verbose_group_info


    # --------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def __split_id_from_version_id(arn_type, combined):
        res = ""
        if arn_type == "core":
            split1 = combined.split("/cores/")
            res = split1[1].split("/versions/")
        if arn_type == "function":
            split1 = combined.split("/functions/")
            res = split1[1].split("/versions/")
        if arn_type == "logger":
            split1 = combined.split("/loggers/")
            res = split1[1].split("/versions/")
        if arn_type == "resources":
            split1 = combined.split("/resources/")
            res = split1[1].split("/versions/")
        if arn_type == "subscriptions":
            split1 = combined.split("/subscriptions/")
            res = split1[1].split("/versions/")
        return res



"""
Created on 28 Sep 2020

@author: Jade Page (jade.page@southcoastscience.com)

https://realpython.com/python-send-email/#option-1-setting-up-a-gmail-account-for-development
"""
import smtplib

from email.mime.multipart import MIMEMultipart

from email.mime.text import MIMEText

from socket import gaierror


# Unused (replaced by AWS SES)

# --------------------------------------------------------------------------------------------------------------------
class EmailClient(object):
    __PORT = 465
    __SMTP_SERVER = "smtp.gmail.com"

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def device_monitor_client(cls, device_monitor_conf):
        email_address = device_monitor_conf.email_name
        email_password = device_monitor_conf.email_password
        return EmailClient(cls.__PORT, cls.__SMTP_SERVER, email_address, email_password)

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, port=0, smtp_server="", login_address="", password=""):
        """
        Constructor
        """
        self.__port = port
        self.__smtp_server = smtp_server
        self.__sender_email = login_address
        self.__password = password
        self.__server = None

    # ----------------------------------------------------------------------------------------------------------------

    def open_server(self):
        # Should this go in the constructor?
        if not self.__port or not self.__smtp_server or not self.__sender_email or not self.__password:
            return "Email Client Not Configured"
        try:
            self.__server = smtplib.SMTP_SSL('%s: %s' % (self.__smtp_server, self.__port))
            self.__server.login(self.__sender_email, self.__password)
        except (gaierror, ConnectionRefusedError):
            return 'Failed to connect to SMTP Server (Refused) '
        except smtplib.SMTPServerDisconnected:
            return 'Failed to connect to SMTP Server (Disconnected) '
        except smtplib.SMTPException as e:
            return 'SMTP error occurred: ' + str(e)
        return "ok"

    def close_server(self):
        if not self.__server:
            return "No server open"
        self.__server.quit()

    def send_mime_email(self, message, device):
        if not self.__server:
            return "No server open"

        msg = MIMEMultipart()
        msg['From'] = self.__sender_email
        msg['To'] = self.__sender_email
        # TODO: Add extra recipients somewhere for each device
        msg['Subject'] = "SCS Device: %s" % device
        msg.attach(MIMEText(message))

        try:
            self.__server.sendmail(msg['From'], msg['To'], msg.as_string())
        except (gaierror, ConnectionRefusedError):
            return 'Failed to connect to SMTP Server (Refused) '
        except smtplib.SMTPServerDisconnected:
            return 'Failed to connect to SMTP Server (Disconnected) '
        except smtplib.SMTPException as e:
            return 'SMTP error occurred: ' + str(e)

        return True

    # ----------------------------------------------------------------------------------------------------------------



"""
Created on 8 Aug 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example:
{"id": "South Coast Science PSU", "tag": "1.0.0", "c-date": "Aug  8 2017", "c-time": "08:35:25"}
"""

from collections import OrderedDict

from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class PSUVersion(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, id, tag_jdict, compile_date=None, compile_time=None):
        tag = PSUTag.construct_from_jdict(tag_jdict)

        return cls(id, tag, compile_date, compile_time)


    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        id = jdict.get('id')

        tag = PSUTag.construct_from_jdict(jdict.get('tag'))

        compile_date = jdict.get('c-date')
        compile_time = jdict.get('c-time')

        return cls(id, tag, compile_date, compile_time)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, id, tag, compile_date, compile_time):
        """
        Constructor
        """
        self.__id = id                                  # string
        self.__tag = tag                                # PSUTag

        self.__compile_date = compile_date              # string (Unix date)
        self.__compile_time = compile_time              # string (time)


    def __eq__(self, other):
        try:
            return self.id == other.id and self.tag == other.tag and \
                   self.compile_date == other.compile_date and self.compile_time == other.compile_time

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['id'] = self.id
        jdict['tag'] = self.tag

        jdict['c-date'] = self.compile_date
        jdict['c-time'] = self.compile_time

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def id(self):
        return self.__id


    @property
    def tag(self):
        return self.__tag


    @property
    def compile_date(self):
        return self.__compile_date


    @property
    def compile_time(self):
        return self.__compile_time


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "PSUVersion:{id:%s, tag:%s, compile_date:%s, compile_time:%s}" \
               % (self.id, self.tag, self.compile_date, self.compile_time)


# --------------------------------------------------------------------------------------------------------------------

class PSUTag(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        items = jdict.split('.')

        device = items[0]
        api = items[1]
        patch = items[2]

        return cls(device, api, patch)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device, api, patch):
        """
        Constructor
        """
        self.__device = device                  # string representation of int
        self.__api = api                        # string representation of int
        self.__patch = patch                    # string representation of int


    def __eq__(self, other):
        try:
            return self.device == other.device and self.api == other.api and self.patch == other.patch

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        return '.'.join((self.device, self.api, self.patch))


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def device(self):
        return self.__device


    @property
    def api(self):
        return self.__api


    @property
    def patch(self):
        return self.__patch


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "PSUTag:{device:%s, api:%s, patch:%s}" % (self.device, self.api, self.patch)



"""
Created on 13 Feb 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

http://www.apis.ac.uk/unit-conversion
https://keisan.casio.com/exec/system/1224579725
"""


# --------------------------------------------------------------------------------------------------------------------

class Gas(object):
    """
    classdocs
    """

    STP_TEMPERATURE =       20.0                                # °C
    STP_PRESSURE =         101.3                                # kPa
    STP_MOLAR_VOLUME =      22.41                               # l

    ST_LAPSE_RATE =          0.0065                             # K/m
    GRAVITY =                9.80665                            # m/s2
    MOLAR_MASS_OF_AIR =     28.9644                             # g/mol
    GAS_CONSTANT =           8.3144598                          # J/(mol*K)


    # ----------------------------------------------------------------------------------------------------------------

    __NAMES = {'CO', 'CO2', 'NO', 'NO2', 'O3', 'SO2'}

    @classmethod
    def is_valid_name(cls, name):
        return name in cls.__NAMES


    # ----------------------------------------------------------------------------------------------------------------

    __MOLAR_DENSITY_STP = {
                'CO':   1.1642,
                'CO2':  1.9600,
                'NO':   1.3402,
                'NO2':  1.9125,
                'O3':   1.9957,
                'SO2':  2.6609
    }

    @classmethod
    def __molar_density_stp(cls, gas):
        if gas not in cls.__MOLAR_DENSITY_STP:
            raise ValueError("Gas: unrecognised name: %s" % gas)

        return cls.__MOLAR_DENSITY_STP[gas]


    __MOLAR_MASS = {
                'CO':   28.0100,
                'CO2':  44.0100,
                'NO':   30.0100,
                'NO2':  46.0055,
                'O3':   48.0000,
                'SO2':  64.0660
    }

    @classmethod
    def __molar_mass(cls, gas):
        if gas not in cls.__MOLAR_MASS:
            raise ValueError("Gas: unrecognised name: %s" % gas)

        return cls.__MOLAR_MASS[gas]


    @classmethod
    def __molar_volume(cls, t, p):
        tk = t + 273.15

        return cls.STP_MOLAR_VOLUME * (tk / 273.15) * (cls.STP_PRESSURE / p)


    # ----------------------------------------------------------------------------------------------------------------
    # density and concentration...

    @classmethod
    def concentration(cls, gas, density, t, p):                 # g/m3, °C, kPa -> ppb
        molar_mass = cls.__molar_mass(gas)
        molar_volume = cls.__molar_volume(t, p)

        molar_density = molar_mass / molar_volume

        return density / molar_density


    @classmethod
    def concentration_stp(cls, gas, density):                   # g/m3 -> ppb
        molar_density = cls.__molar_density_stp(gas)

        return density / molar_density


    @classmethod
    def density(cls, gas, concentration, t, p):                 # ppb, °C, kPa -> g/m3
        molar_mass = cls.__molar_mass(gas)
        molar_volume = cls.__molar_volume(t, p)

        molar_density = molar_mass / molar_volume

        return concentration * molar_density


    @classmethod
    def density_stp(cls, gas, concentration):                   # ppb -> g/m3
        molar_density = cls.__molar_density_stp(gas)

        return concentration * molar_density


    # ----------------------------------------------------------------------------------------------------------------
    # pressure...

    @classmethod
    def p_alt(cls, p_0, t, alt):                                # KPa, °C, m -> KPa
        return p_0 * cls.__hypsometric(t, alt)


    @classmethod
    def p_0(cls, p_alt, t, alt):                                # KPa, °C, m -> KPa
        return p_alt / cls.__hypsometric(t, alt)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __hypsometric(cls, t, alt):
        tk = t + 273.15
        lapse = alt * cls.ST_LAPSE_RATE
        a = (cls.GRAVITY * (cls.MOLAR_MASS_OF_AIR / 1000)) / (cls.GAS_CONSTANT * cls.ST_LAPSE_RATE)

        return pow(1 - (lapse / (tk + lapse)), a)



"""
Created on 8 Sep 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example JSON:
{"sample-interval": 2, "temp-offset": 0.0}
"""

from collections import OrderedDict

from scs_core.data.datum import Datum
from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class SCD30Conf(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "scd30_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        sample_interval = jdict.get('sample-interval')
        temperature_offset = jdict.get('temp-offset')

        return cls(sample_interval, temperature_offset)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, sample_interval, temperature_offset):
        """
        Constructor
        """
        super().__init__()

        self.__sample_interval = Datum.int(sample_interval)                     # int       seconds
        self.__temperature_offset = Datum.float(temperature_offset, 1)          # float     °C


    def __eq__(self, other):
        try:
            return self.sample_interval == other.sample_interval and \
                   self.temperature_offset == other.temperature_offset

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def scd30():
        return None


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['sample-interval'] = self.sample_interval
        jdict['temp-offset'] = self.temperature_offset

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def sample_interval(self):
        return self.__sample_interval


    @property
    def temperature_offset(self):
        return self.__temperature_offset


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SCD30Conf(core):{sample_interval:%s, temperature_offset:%s}" %  \
               (self.sample_interval, self.temperature_offset)



"""
Created on 20 Jan 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)


https://stackoverflow.com/questions/16175745/read-new-line-with-pynotify
https://www.linode.com/docs/development/monitor-filesystem-events-with-pyinotify/
"""

import sys

from pyinotify import EventsCodes, Notifier, ProcessEvent, WatchManager


# --------------------------------------------------------------------------------------------------------------------

class Tail(object):
    """
    classdocs
    """

    DEFAULT_TIMEOUT =   600000                  # 10 minutes in milliseconds

    # ----------------------------------------------------------------------------------------------------------------

    __IN_MODIFY =       EventsCodes.FLAG_COLLECTIONS['OP_FLAGS']['IN_MODIFY']
    __IN_CLOSE_WRITE =  EventsCodes.FLAG_COLLECTIONS['OP_FLAGS']['IN_CLOSE_WRITE']
    __IN_MOVED_TO =     EventsCodes.FLAG_COLLECTIONS['OP_FLAGS']['IN_MOVED_TO']

    __IN_EVENTS =       __IN_MODIFY | __IN_CLOSE_WRITE | __IN_MOVED_TO


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, path, timeout=None):
        handler = TailEventHandler(path)

        watch_manager = WatchManager()
        watch_manager.add_watch(path, cls.__IN_EVENTS)

        notifier_timeout = cls.DEFAULT_TIMEOUT if timeout is None else timeout
        notifier = TailNotifier(watch_manager, default_proc_fun=handler, timeout=notifier_timeout)

        return cls(notifier, handler)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, notifier, handler):
        """
        Constructor
        """
        self.__notifier = notifier                              # TailNotifier
        self.__handler = handler                                # TailEventHandler


    def __iter__(self):
        return self.readlines()


    # ----------------------------------------------------------------------------------------------------------------

    def open(self):
        self.__handler.open_file()


    def close(self):
        self.__handler.close_file()


    def readlines(self):
        # existing lines...
        for line in self.__handler.read_head():
            yield line

        # new lines...
        for line in self.__notifier.loop(self.__handler.read_tail):
            yield line

        # raise StopIteration


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Tail:{handler:%s}" %  self.__handler


# --------------------------------------------------------------------------------------------------------------------

class TailEventHandler(ProcessEvent):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, path):
        """
        Constructor
        """
        super().__init__()

        self.__path = path

        self.__file = None
        self.__terminate = False


    # ----------------------------------------------------------------------------------------------------------------

    def open_file(self):
        self.__file = open(self.__path)


    def close_file(self):
        if self.__file is None:
            return

        self.__file.close()
        self.__file = None


    def read_head(self, _notifier=None):
        for line in self.__file.readlines():
            yield line.strip()


    def read_tail(self, _notifier=None):
        if self.__terminate:
            return True

        return [line.strip() for line in self.__file.readlines()]


    # ----------------------------------------------------------------------------------------------------------------

    # noinspection PyPep8Naming
    def process_IN_MODIFY(self, _event):
        # self.__report("IN_MODIFY")
        pass


    # noinspection PyPep8Naming
    def process_IN_CLOSE_WRITE(self, _event):
        self.__report("IN_CLOSE_WRITE")

        self.__terminate = True


    # noinspection PyPep8Naming
    def process_IN_MOVED_TO(self, _event):
        self.__report("IN_MOVED_TO")

        self.__terminate = True


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __report(cls, event_name):
        print("TailEventHandler: %s" % event_name, file=sys.stderr)
        sys.stderr.flush()


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "TailEventHandler:{path:%s, terminate:%s}" % (self.__path, self.__terminate)


# --------------------------------------------------------------------------------------------------------------------

class TailNotifier(Notifier):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, watch_manager, default_proc_fun=None, read_freq=0, threshold=0, timeout=None):
        super().__init__(watch_manager, default_proc_fun, read_freq, threshold, timeout)


    # ----------------------------------------------------------------------------------------------------------------

    def loop(self, callback=None, _daemonize=False, **_args):
        try:
            while True:
                self.process_events()

                response = callback(self)

                if response is True:
                    break                                   # file closed

                for line in response:
                    yield line

                if not self.check_events():
                    raise TimeoutError                      # timeout

                self.read_events()

        except (ConnectionError, KeyboardInterrupt, SystemExit):
            pass

        finally:
            self.stop()



"""
Created on 11 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

NB: initialisation is performed at the foot of this class

example document:
{"topic": "particulates", "species": "pm2p5", "source": "N3", "parameter-code": "88101", "unit-code": 105,
"qc-code": 0, "method-code": 195, "mpc-code": 1, "mpc-value": 5.0}
"""

import os

from collections import OrderedDict

from scs_core.csv.csv_archive import CSVArchive

from scs_core.data.datum import Datum
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class SourceMapping(CSVArchive, JSONable):
    """
    classdocs
    """

    _retrieved = {}

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def archive_location(cls):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)), 'archive', 'source_mappings.csv')


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if jdict is None:
            return None

        topic = jdict.get('topic')
        species = jdict.get('species')
        source = jdict.get('source')

        parameter_code = jdict.get('parameter-code')
        unit_code = jdict.get('unit-code')
        qc_code = jdict.get('qc-code')
        method_code = jdict.get('method-code')
        mpc_code = jdict.get('mpc-code')

        mpc_value = jdict.get('mpc-value')

        return SourceMapping(topic, species, source,
                             parameter_code, unit_code, qc_code, method_code, mpc_code, mpc_value)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, topic, species, source,
                 parameter_code, unit_code, qc_code, method_code, mpc_code, mpc_value):
        """
        Constructor
        """
        self.__topic = topic                                        # string
        self.__species = species                                    # string
        self.__source = source                                      # string

        self.__parameter_code = parameter_code                      # int(5) or string
        self.__unit_code = Datum.int(unit_code)                     # int(3)
        self.__qc_code = Datum.int(qc_code)                         # int
        self.__method_code = Datum.int(method_code)                 # int(3)
        self.__mpc_code = Datum.int(mpc_code)                       # int

        self.__mpc_value = Datum.float(mpc_value, 5)                # numeric(10,5)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['topic'] = self.topic
        jdict['species'] = self.species
        jdict['source'] = self.source

        jdict['parameter-code'] = self.parameter_code
        jdict['unit-code'] = self.unit_code
        jdict['qc-code'] = self.qc_code
        jdict['method-code'] = self.method_code
        jdict['mpc-code'] = self.mpc_code

        jdict['mpc-value'] = self.mpc_value

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def pk(self):
        return self.topic, self.species, self.source


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def topic(self):
        return self.__topic


    @property
    def species(self):
        return self.__species


    @property
    def source(self):
        return self.__source


    @property
    def parameter_code(self):
        return self.__parameter_code


    @property
    def unit_code(self):
        return self.__unit_code


    @property
    def qc_code(self):
        return self.__qc_code


    @property
    def method_code(self):
        return self.__method_code


    @property
    def mpc_code(self):
        return self.__mpc_code


    @property
    def mpc_value(self):
        return self.__mpc_value


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SourceMapping:{topic:%s, species:%s, source:%s, " \
               "parameter_code:%s, unit_code:%03d, qc_code:%d, method_code:%03d, mpc_code:%d, mpc_value:%0.5f}" % \
               (self.topic, self.species, self.source,
                self.parameter_code, self.unit_code, self.qc_code, self.method_code, self.mpc_code, self.mpc_value)


# --------------------------------------------------------------------------------------------------------------------
# initialisation...

SourceMapping.retrieve()



"""
Created on 10 Jan 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://www.nmea.org
https://en.wikipedia.org/wiki/NMEA_0183

reference coordinate systems:
https://en.wikipedia.org/wiki/World_Geodetic_System#WGS84
https://en.wikipedia.org/wiki/PZ-90
"""

from collections import OrderedDict
from numbers import Number

from scs_core.data.datum import Datum
from scs_core.data.json import JSONReport

from scs_core.position.position import Position


# --------------------------------------------------------------------------------------------------------------------

class GPSDatum(JSONReport):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls.null_datum()

        pos = Position.construct_from_jdict(jdict.get('pos'))
        elv = jdict.get('elv')

        quality = jdict.get('qual')

        return GPSDatum(pos, elv, quality)


    @classmethod
    def construct_from_gga(cls, gga):
        if gga is None:
            return None

        pos = Position.construct_from_gga(gga)
        elv = None if gga.alt is None else round(gga.alt)

        quality = gga.quality

        return GPSDatum(pos, elv, quality)


    @classmethod
    def null_datum(cls):
        return GPSDatum(Position(None, None), None, None)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, pos, elv, quality):
        """
        Constructor
        """
        self.__pos = pos                            # Position
        self.__elv = Datum.float(elv, 1)            # metres above mean sea level

        self.__quality = quality                    # number or None


    # ----------------------------------------------------------------------------------------------------------------
    # Support for averaging...

    def __add__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError(other)

        pos = self.pos + other.pos
        elv = self.elv + other.elv

        quality = self.quality + other.quality

        return GPSDatum(pos, elv, quality)


    def __truediv__(self, other):
        if not isinstance(other, Number):
            raise TypeError(other)

        pos = self.pos / other
        elv = self.elv / other

        quality = self.quality / other

        return GPSDatum(pos, elv, quality)


    # ----------------------------------------------------------------------------------------------------------------

    def distance(self, other_pos, minimum_acceptable_quality=None):
        if self.pos is None:
            return None

        if minimum_acceptable_quality is not None:
            if self.quality is None or round(self.quality) < minimum_acceptable_quality:
                return None

        return self.pos.distance(other_pos)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['pos'] = self.pos
        jdict['elv'] = None if self.elv is None else round(self.elv, 1)

        jdict['qual'] = None if self.quality is None else round(self.quality, 1)

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def pos(self):
        return self.__pos


    @property
    def elv(self):
        return self.__elv


    @property
    def quality(self):
        return self.__quality


    @quality.setter
    def quality(self, quality):
        self.__quality = quality


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "GPSDatum:{pos:%s, elv:%s, quality:%s}" % (self.pos, self.elv, self.quality)



"""
Created on 9 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

header:
CURLOPT_HTTPHEADER => array('Accept: application/json', 'Authorization: api-key ' . AOC_OPENSENSORS_API_KEY),
"""

import json

from scs_core.data.json import JSONify

from scs_core.client.http_client import HTTPClient

from scs_core.osio.client.client_exception import ClientException

from scs_core.sys.http_exception import HTTPException
from scs_core.sys.http_status import HTTPStatus


# --------------------------------------------------------------------------------------------------------------------

class RESTClient(object):
    """
    classdocs
    """

    __HOST = "api.opensensors.io"               # hard-coded URL

    __VERIFIED = False                          # False - ignore invalid SSL certificates

    __HEADER_ACCEPT = "application/json"
    __HEADER_AUTHORIZATION = "api-key "


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, api_key):
        """
        Constructor
        """
        self.__http_client = HTTPClient()
        self.__api_key = api_key


    # ----------------------------------------------------------------------------------------------------------------

    def connect(self, timeout=None):
        self.__http_client.connect(self.__HOST, verified=self.__VERIFIED, timeout=timeout)


    def close(self):
        self.__http_client.close()


    # ----------------------------------------------------------------------------------------------------------------

    def get(self, path, params=None):
        try:
            response_jstr = self.__http_client.get(path, params, self.__headers)
        except HTTPException as ex:
            if ex.status == HTTPStatus.NOT_FOUND:
                return None
            else:
                raise ClientException.construct(ex) from ex

        try:
            response = json.loads(response_jstr)
        except ValueError:
            response = None

        return response


    def post(self, path, payload_jdict):                # TODO: make the jdict here?
        payload_jstr = JSONify.dumps(payload_jdict)

        # print("RESTClient.post: path: %s payload: %s" % (path, payload_jstr))

        try:
            response_jstr = self.__http_client.post(path, payload_jstr, self.__headers)
        except HTTPException as ex:
            raise ClientException.construct(ex) from ex

        try:
            response = json.loads(response_jstr)
        except ValueError:
            response = None

        return response


    def put(self, path, payload_jdict):                # TODO: make the jdict here?
        payload_jstr = JSONify.dumps(payload_jdict)

        # print("RESTClient.put: path: %s payload: %s" % (path, payload_jstr))

        try:
            response_jstr = self.__http_client.put(path, payload_jstr, self.__headers)
        except HTTPException as ex:
            raise ClientException.construct(ex) from ex

        try:
            response = json.loads(response_jstr)
        except ValueError:
            response = None

        return response


    def delete(self, path):
        try:
            response_jstr = self.__http_client.delete(path, self.__headers)
        except HTTPException as ex:
            raise ClientException.construct(ex) from ex

        try:
            response = json.loads(response_jstr)
        except ValueError:
            response = None

        return response


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def __headers(self):
        return {"Accept": self.__HEADER_ACCEPT,
                "Authorization": self.__HEADER_AUTHORIZATION + self.__api_key}


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "RESTClient:{http_client:%s, api_key:%s}" % (self.__http_client, self.__api_key)



"""
Created on 24 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://docs.python.org/3/library/struct.html
https://stackoverflow.com/questions/8022530/how-to-check-for-valid-email-address
"""

import math
import re
import struct

from datetime import date


# --------------------------------------------------------------------------------------------------------------------

class Datum(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------
    # formatting...

    @classmethod
    def format(cls, value, leading_zeroes=False):
        # string or bool...
        if not cls.is_numeric(value):
            return "%s"

        # numeric...
        prefix = "0" if leading_zeroes else ""
        length = str(len((str(value))))

        if cls.is_int(value):
            return "%" + prefix + length + "d"

        # float...
        precision = str(cls.precision(value))

        return "%" + prefix + length + "." + precision + "f"


    # ----------------------------------------------------------------------------------------------------------------
    # morphological numeracy...

    @classmethod
    def is_email_address(cls, value):
        try:
            return bool(re.match(r'[^@\s]+@[^@\s]+\.[^@\s]+', value))
        except TypeError:
            return False


    @classmethod
    def is_numeric(cls, value):
        return cls.precision(value) is not None


    @classmethod
    def is_int(cls, value):
        precision = cls.precision(value)

        if precision is None:
            return False

        return precision == 0


    @classmethod
    def is_float(cls, value):
        precision = cls.precision(value)

        if precision is None:
            return False

        return precision > 0


    @staticmethod
    def precision(value):
        if value is None:
            return None

        if isinstance(value, bool):
            return None

        try:
            float(value)
        except ValueError:
            return None

        pieces = str(value).split('.')

        # int...
        if len(pieces) == 1:
            return 0                            # warning: round(123, 0) returns 123.0 - use round(123)

        # float...
        return len(pieces[1])                   # warning: interprets 1. as precision 0


    # ----------------------------------------------------------------------------------------------------------------
    # cast or None...

    @staticmethod
    def bool(field):
        if field is None:
            return None

        try:
            value = bool(field)
        except ValueError:
            return None

        return value


    @staticmethod
    def int(number):
        if number is None:
            return None

        try:
            value = int(number)
        except ValueError:
            return None

        return int(value)


    @staticmethod
    def float(number, ndigits=None):
        if number is None:
            return None

        try:
            value = float(number)
        except ValueError:
            return None

        return value if ndigits is None else round(value, ndigits)      # warning: round(123, 0) returns 123.0


    @staticmethod
    def date(iso_date):
        if iso_date is None:
            return None

        parts = iso_date.split("-")

        if len(parts) != 3:
            return None

        try:
            year = int(float(parts[0]))
            month = int(float(parts[1]))
            day = int(float(parts[2]))
        except ValueError:
            return None

        return date(year, month, day)


# --------------------------------------------------------------------------------------------------------------------

class Encode(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------
    # encode byte array...

    @staticmethod
    def bool(value):
        return 1 if value else 0


    @staticmethod
    def int(value, order):
        unpacked = struct.unpack('BB', struct.pack(order + 'h', int(value)))

        return unpacked


    @staticmethod
    def unsigned_int(value, order):
        unpacked = struct.unpack('BB', struct.pack(order + 'H', int(value)))

        return unpacked


    @staticmethod
    def unsigned_long(value, order):
        unpacked = struct.unpack('BBBB', struct.pack(order + 'L', int(value)))

        return unpacked


    @staticmethod
    def float(value, order):
        unpacked = struct.unpack('BBBB', struct.pack(order + 'f', float(value)))

        return unpacked


# --------------------------------------------------------------------------------------------------------------------

class Decode(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------
    # decode byte array...

    @staticmethod
    def int(byte_values, order):
        packed = struct.unpack(order + 'h', struct.pack('BB', *byte_values))

        return packed[0]


    @staticmethod
    def unsigned_int(byte_values, order):
        packed = struct.unpack(order + 'H', struct.pack('BB', *byte_values))

        return packed[0]


    @staticmethod
    def long(byte_values, order):
        packed = struct.unpack(order + 'l', struct.pack('BBBB', *byte_values))

        return packed[0]


    @staticmethod
    def unsigned_long(byte_values, order):
        packed = struct.unpack(order + 'L', struct.pack('BBBB', *byte_values))

        return packed[0]


    @staticmethod
    def float(byte_values, order):
        packed = struct.unpack(order + 'f', struct.pack('BBBB', *byte_values))

        return None if math.isnan(packed[0]) else packed[0]


    @staticmethod
    def double(byte_values, order):
        packed = struct.unpack(order + 'd', struct.pack('BBBBBBBB', *byte_values))

        return None if math.isnan(packed[0]) else packed[0]



"""
Created on 20 Oct 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import os
import time
import tzlocal

from abc import ABC, abstractmethod

from scs_core.data.datetime import LocalizedDatetime
from scs_core.sys.filesystem import Filesystem


# --------------------------------------------------------------------------------------------------------------------

class PersistenceManager(ABC):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def list(self, container, dirname):
        pass


    @abstractmethod
    def exists(self, dirname, filename):
        pass


    @abstractmethod
    def load(self, dirname, filename, encryption_key=None):
        pass


    @abstractmethod
    def save(self, jstr, dirname, filename, encryption_key=None):
        pass


    @abstractmethod
    def remove(self, dirname, filename):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def scs_path(cls):
        pass


# --------------------------------------------------------------------------------------------------------------------

class FilesystemPersistenceManager(PersistenceManager, ABC):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def list(cls, container, dirname):
        abs_dirname = os.path.join(container, dirname)

        return sorted(os.listdir(abs_dirname))


    @classmethod
    def exists(cls, dirname, filename):
        abs_filename = cls.__abs_filename(dirname, filename)

        return os.path.isfile(abs_filename)


    @classmethod
    def load(cls, dirname, filename, encryption_key=None):
        abs_filename = cls.__abs_filename(dirname, filename)

        try:
            with open(abs_filename, "r") as f:
                text = f.read()

            if encryption_key:
                from scs_core.data.crypt import Crypt               # late import
                jstr = Crypt.decrypt(encryption_key, text)
            else:
                jstr = text

        except FileNotFoundError:
            return None, None

        timestamp = int(os.path.getmtime(abs_filename))
        last_modified = LocalizedDatetime.construct_from_timestamp(timestamp, tz=tzlocal.get_localzone()).utc()

        return jstr, last_modified


    @classmethod
    def save(cls, jstr, dirname, filename, encryption_key=None):
        abs_dirname = cls.__abs_dirname(dirname)
        abs_filename = cls.__abs_filename(dirname, filename)

        # file...
        if filename:
            Filesystem.mkdir(abs_dirname)

        tmp_filename = '.'.join((abs_filename, str(int(time.time()))))

        if encryption_key:
            from scs_core.data.crypt import Crypt                   # late import
            text = Crypt.encrypt(encryption_key, jstr)
        else:
            text = jstr + '\n'

        with open(tmp_filename, "w") as f:
            f.write(text)

        # atomic operation...
        os.rename(tmp_filename, abs_filename)


    @classmethod
    def remove(cls, dirname, filename):
        abs_filename = cls.__abs_filename(dirname, filename)

        try:
            os.remove(abs_filename)
        except FileNotFoundError:
            pass


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __abs_filename(cls, dirname, filename):
        return os.path.join(cls.scs_path(), dirname, filename)


    @classmethod
    def __abs_dirname(cls, dirname):
        return os.path.join(cls.scs_path(), dirname)



"""
Created on 13 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import sys
import time

import urllib.parse

from scs_core.data.datetime import LocalizedDatetime

from scs_core.osio.client.rest_client import RESTClient
from scs_core.osio.data.message import Message


# --------------------------------------------------------------------------------------------------------------------

class MessageManager(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, api_key, verbose=False):
        """
        Constructor
        """
        self.__rest_client = RESTClient(api_key)
        self.__verbose = verbose


    # ----------------------------------------------------------------------------------------------------------------

    def find_for_topic(self, topic, start_date, end_date, batch_pause=0.0):
        request_path = '/v1/messages/topic/' + topic

        total = 0
        collection = []

        # request...
        self.__rest_client.connect()

        try:
            for batch in self.__find(request_path, start_date, end_date):
                collection.extend(batch)

                if self.__verbose:
                    now = LocalizedDatetime.now().utc()
                    batch_count = len(batch)
                    total += batch_count

                    print("%s: batch: %d total: %d" % (now.as_iso8601(), batch_count, total), file=sys.stderr)
                    sys.stderr.flush()

                time.sleep(batch_pause)     # prevent "Rate limit exceeded" error

        finally:
            self.__rest_client.close()

        return collection


    # ----------------------------------------------------------------------------------------------------------------

    def __find(self, request_path, start_date, end_date):
        params = {'start-date': start_date.as_iso8601(), 'end-date': end_date.as_iso8601()}

        while True:
            # request...
            jdict = self.__rest_client.get(request_path, params=params)

            # messages...
            msgs_jdict = jdict.get('messages')
            messages = [Message.construct_from_jdict(msg_jdict) for msg_jdict in msgs_jdict] if msgs_jdict else []

            yield messages

            # next...
            next_jdict = jdict.get('next')
            next_query = NextMessageQuery.construct_from_uri(next_jdict)

            if next_query is None:
                return

            params['start-date'] = next_query.start_date.as_iso8601()
            params['end-date'] = next_query.end_date.as_iso8601()


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MessageManager:{rest_client:%s, verbose:%s}" % (self.__rest_client, self.__verbose)


# --------------------------------------------------------------------------------------------------------------------

class NextMessageQuery(object):
    """
    classdocs

    example query:
    /v1/messages/topic//users/southcoastscience-dev/test/json?
    start-date=2016-11-13T07:11:14.779Z&end-date=2016-11-13T08:48:08.901+00:00
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_uri(cls, uri):
        if not uri:
            return None

        # parse...
        parse = urllib.parse.urlparse(urllib.parse.unquote(uri))
        params = urllib.parse.parse_qs(parse[4])

        if 'start-date' not in params or 'end-date' not in params:
            return None

        # construct...
        start_date = LocalizedDatetime.construct_from_iso8601(params['start-date'][0])
        end_date = LocalizedDatetime.construct_from_iso8601(params['end-date'][0])

        if not start_date or not end_date:
            return None

        return NextMessageQuery(start_date, end_date)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, start_date, end_date):
        """
        Constructor
        """
        self.__start_date = start_date          # LocalizedDatetime
        self.__end_date = end_date              # LocalizedDatetime


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def start_date(self):
        return self.__start_date


    @property
    def end_date(self):
        return self.__end_date


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "NextMessageQuery:{start_date:%s, end_date:%s}" % (self.start_date, self.end_date)



"""
Created on 13 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

"""

from scs_core.osio.client.rest_client import RESTClient
from scs_core.osio.data.device import Device
from scs_core.osio.data.device_summary import DeviceSummary


# --------------------------------------------------------------------------------------------------------------------

class DeviceManager(object):
    """
    classdocs
    """
    __FINDER_BATCH_SIZE = 100


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, api_key):
        """
        Constructor
        """
        self.__rest_client = RESTClient(api_key)


    # ----------------------------------------------------------------------------------------------------------------

    def find(self, org_id, client_id):
        request_path = '/v1/orgs/' + org_id + '/devices/' + client_id

        # request...
        self.__rest_client.connect()

        try:
            response_jdict = self.__rest_client.get(request_path)
        except RuntimeError:
            response_jdict = None

        self.__rest_client.close()

        device = Device.construct_from_jdict(response_jdict)

        return device


    def find_for_name(self, org_id, name):
        devices = self.find_all_for_org(org_id)

        for device in devices:                              # warning: unique only by convention
            if device.name == name:
                return self.find(org_id, device.client_id)

        return None


    def find_all_for_user(self, user_id):
        request_path = '/v1/users/' + user_id + '/devices'
        devices = []

        # request...
        self.__rest_client.connect()

        try:
            for batch in self.__find(request_path, {'user-id': user_id}):
                devices.extend(batch)

        finally:
            self.__rest_client.close()

        return devices


    def find_all_for_org(self, org_id):
        request_path = '/v1/orgs/' + org_id + '/devices'
        devices = []

        # request...
        self.__rest_client.connect()

        try:
            for batch in self.__find(request_path):
                devices.extend(batch)

        finally:
            self.__rest_client.close()

        return devices


    # ----------------------------------------------------------------------------------------------------------------

    def __find(self, request_path, params=None):
        if params is None:
            params = {}

        params['offset'] = 0
        params['count'] = self.__FINDER_BATCH_SIZE

        while True:
            # request...
            response_jdict = self.__rest_client.get(request_path, params=params)

            devices = [DeviceSummary.construct_from_jdict(jdict) for jdict in response_jdict] if response_jdict else []

            yield devices

            if len(devices) == 0:
                break

            # next...
            params['offset'] += len(devices)


    # ----------------------------------------------------------------------------------------------------------------

    def create(self, user_id, device):
        request_path = '/v1/users/' + user_id + '/devices'

        # request...
        self.__rest_client.connect()

        try:
            response_jdict = self.__rest_client.post(request_path, device.as_json())
        finally:
            self.__rest_client.close()

        device = Device.construct_from_jdict(response_jdict)

        return device


    def update(self, org_id, system_id, device):
        request_path = '/v1/orgs/' + org_id + '/devices/' + system_id

        # request...
        self.__rest_client.connect()

        try:
            self.__rest_client.put(request_path, device.as_json())
        finally:
            self.__rest_client.close()


    def delete(self, user_id, client_id):
        request_path = '/v1/users/' + user_id + '/devices/' + client_id

        # request...
        self.__rest_client.connect()

        try:
            self.__rest_client.delete(request_path)
        finally:
            self.__rest_client.close()


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "DeviceManager:{rest_client:%s}" % self.__rest_client
